//==================================================================
// Dom::ShiftOperators - Shift Operators
//------------------------------------------------------------------
// blah...
//------------------------------------------------------------------

// for convenience
alias (rep(p) = extop(p,1));
                      
domain Dom::ShiftOperators(X=x, Shift=S, F=Dom::ExpressionField())

local
    R,
    P; // Polynomial Ring for internal rep.
       
inherits Dom::BaseDomain; //TruncatedPowerSeries();//[x, xdx], F);
category Cat::Ring;

BaseField := F;
Indet := X;
characteristic := F::characteristic;

//make_slot := proc(d,s) begin print(Unquoted, "Slot requested: ".expr2text(s)); FAIL; end_proc;

whoami := () -> "Dom::ShiftOperators";

//==================================================================
// Creating new Elements
//------------------------------------------------------------------
// new
//------------------------------------------------------------------
new := proc(p) 
local i,d,r,t,pp,dom2;
save f; 
begin
  delete f; 
  if p = FAIL then return(FAIL) end_if;
  dom2 := domtype(p);
  if dom2 = dom then
    return(p)
  elif dom2 = R then 
    if not has(expr(p), Shift) then
      p := hold(_mult)(1,p);
    else
      // strange, uh? Necessary for ExpressionFields
      p := expr(p);
    end_if;
  elif dom2 = DOM_POLY then
    p := expr(p)
  elif dom2 = DOM_EXPR then
    p := expr(p)
  elif domtype(dom2) = DOM_DOMAIN and dom2::whoami() = dom::whoami() then
    t := rep(p);
    if op(t,2) = [Shift] and dom2::Indet = X then
      t := poly(expr(t),[Shift],R);
      if t <> FAIL then
        return(new(dom, t));
      end_if;
    end_if;
    return(FAIL);
  elif domtype(dom2) = DOM_DOMAIN then
    p := expr(p)
  end_if;
  
  pp := collect(p, Shift, f);
  d := degree(pp, Shift);
  r := [[R(op(coeff(pp,Shift,i))), i] $ i=0..d];
  if has(r, FAIL) then error("Can't convert to ShiftOp"); return(FAIL); end_if;
  r := poly(r, [Shift], R);
  new(dom, r);
end_proc; // new

//------------------------------------------------------------------
// random
//------------------------------------------------------------------
random := proc()
begin
  FAIL;
  //dom::new(X^mods(random(), 7)*polylib::randpoly([XDX], Monic))
end; // random

//------------------------------------------------------------------
// coerce
//------------------------------------------------------------------
coerce := dom::new;

// Dunno
convert := () -> FAIL;

//==================================================================
// Constants
//------------------------------------------------------------------
one := dom::new(1); 
zero := dom::new(0); 
characteristic := 0;

//==================================================================
// Arithmetics
//------------------------------------------------------------------
// _plus -- Add two or more EuDOs
//------------------------------------------------------------------
_plus := proc()
local a, t, f1, f2, A;
begin
  A := [ _if(domtype(a) = dom, a, _if((t := dom::new(a)) <> FAIL, t, a)) $ a in args()];
  A := split(A, (a) -> domtype(a)=dom);
  f1 := dom::_plus2(op(A[1]));
  if (nops(A[2])+nops(A[3])>0) then
    f2 := _plus(op(A[2].A[3]));
    return(hold(_plus)(f1,f2));
  else
    return(f1);
  end_if;
end_proc; // _plus

_plus2 := proc(p, q) local i; 
begin
  if args(0) = 1 then
    return(p)
  end_if;
  if (args(0) > 2) then
    q := dom::_plus2(args(i) $ i=2..args(0));
  end_if;
  return(new(dom, rep(p)+rep(q)));
end_proc; // _plus2

//------------------------------------------------------------------
// _negate -- get -L from L
//------------------------------------------------------------------
_negate := (p) -> dom::new(-1)*p;

//------------------------------------------------------------------
// _mult --  Multiply two or more Shift Operators
//------------------------------------------------------------------
_mult := proc()
local a, t, f1, f2, A;
begin
  A := [ _if(domtype(a) = dom, a, _if((t := dom::new(a)) <> FAIL, t, a)) $ a in args()];
  A := split(A, (a) -> domtype(a)=dom);
  f1 := dom::_mult2(op(A[1]));
  if (nops(A[2])+nops(A[3])>0) then
    f2 := _mult(op(A[2].A[3]));
    return(hold(_mult)(f1,f2));
  else
    return(f1);
  end_if;
end_proc; // _mult

_mult2 := proc(p:dom, q:dom) 
  local r, l1, l2, e1, c1, e2, c2, t, i, j;
begin
  if args(0) = 0 then
    return(dom::one);
  elif args(0) = 1 then
    return(p);
  end_if;
  if args(0) > 2 then
    // though not commutative, it's associative
    q := _mult(args(i) $ i=2..args(0));
  end_if; 
  r := dom::zero;
  l1 := poly2list(rep(p));
  l2 := poly2list(rep(q));
  for i in l1 do
    e1 := i[2]; // exponent of S in p
    c1 := i[1]; // coefficient of S^e1 in p
    for j in l2 do
      e2 := j[2]; // exponent of S in q
      c2 := j[1]; // coefficient of S^e2 in q
      // now, we must compute (S^e1)*c2 
      t := dom::Commute(e1, c2);
      // special case for the lclm
      r := r + new(dom, poly([[R(c1*t),e2+e1]], [Shift], R));
    end_for;
  end_for;
  return(dom::new(r));
end_proc; // _mult2

//------------------------------------------------------------------
// Commute -- compute the commuted expression 
// Shift^e * c = Commute(e, c) * Shift^e
//------------------------------------------------------------------
Commute := proc(e : DOM_INT, c)
local n,d,nr,dr,t, cp, i, mb;
//option remember;
begin
  mb := dom::mybinom;
  cp := (m, co) -> 
    poly([[co*mb(m,i)*e^(m-i),i] $ i=0..m],[X]); 
  n := poly2list(R::numer(c));
  d := poly2list(R::denom(c));
  nr := _plus( cp(t[2], t[1]) $ t in n);
  dr := _plus( cp(t[2], t[1]) $ t in d);
  return(R([nr,dr]));
end_proc; // Commute

//------------------------------------------------------------------
// _divide -- rightdivide p1 by p2
//             i.e., determine q and r such that
//             p1 = p2*q + r
//             where r is of lower or equal degree than p2
//------------------------------------------------------------------
_divide := dom::rightdivide;
rightdivide := proc(p1 : dom, p2 : dom)
local r,q,p,t,lcp, lcr, lcq,dd;
begin
  //return (FAIL);
  r := p1; // remainder
  q := dom::zero;  // quotient
  p := p2;
  while r <> dom::zero do 
    dd := degree(r, Shift)-degree(p, Shift);
    if dd < 0 then break end_if;
    [lcp, lcr] := [lcoeff(p, Shift), lcoeff(r, Shift)]; 
    lcp := R::shift(lcp, dd); // shift back
//    lcp := R::shift(expr(lcp) | X=X+dd); // shift back
    lcq := lcr/lcp;
    t := dom::new(lcq) * dom::new(Shift^dd);
    q := q + t;
    r := r - t*p;
  end_while;
  return ([q,r]);
end_proc; // rightdivide

//------------------------------------------------------------------
// leftdivide -- leftdivide p1 by p2
//                i.e., determine q and r such that
//                p1 = q*p2 + r
//                where r is of lower or equal degree than p2
//------------------------------------------------------------------
leftdivide := proc(p1 : dom, p2 : dom)
local r,q,p,t,d,lcp, lcr, lcq,dd;
begin
  r := p1; // remainder
  q := dom::zero;  // quotient
  p := p2;
  while r <> dom::zero do //and FAIL <> ((t := (poly(lmonomial(r))/poly(lmonomial(p))))) do
    dd := degree(r, Shift)-degree(p, Shift);
    if dd < 0 then break end_if;
    d := degree(p, Shift);
    [lcp, lcr] := [lcoeff(p, Shift), lcoeff(r, Shift)]; 
    lcq := lcr/lcp;
    lcq := R::shift(lcq, -d); // shift back
//    lcq := R(expr(lcq) | X=X-d); // shift back
    t := dom::new(lcq) * dom::new(Shift^dd);
    q := q + t;
    r := r - p*t;
  end_while;
  return ([q,r]);
end_proc; // leftdivide

//------------------------------------------------------------------
// gcrd -- greatest common right divisor
//------------------------------------------------------------------
gcd := dom::gcrd;
gcrd := proc(p:dom, q:dom) 
local t, f1, f2;
begin
  // monify input
  p := dom::new((lcoeff(p))^(-1))*p;
  q := dom::new((lcoeff(q))^(-1))*q;    
  
  [f1, f2] := _if(degree(p)>degree(q),[p,q],[q,p]);
  while TRUE do // exit via break
    [q, t] := f1/f2;
    if dom::iszero(t) then break end_if; 
    [f1, f2] := [f2, dom::new((lcoeff(t))^(-1))*t];
  end_while;
  return (f2);
end_proc; // gcrd

//------------------------------------------------------------------
// lclm -- least common left multiple
//------------------------------------------------------------------
lcm := dom::lclm;
lclm := proc(p:dom, q:dom) 
local n,i,P,Q, E, AA,conv,Ansatz,Sol, MS, d, q1, q2, r1, r2;
begin
  n := degree(p) + degree(q);
  // to allow the indets in the operator, we mus enlarge the field
  MS := Dom::ShiftOperators(X, Shift, Dom::ExpressionField());
  P := MS(expr(p));
  Q := MS(expr(q));
  repeat
    // now, create an SO which has only variables as coefficients
    AA := [`#`.i $ i=0..n-1];
    Ansatz := MS(Shift^n+_plus((AA[i+1] * Shift^i) $ i=0..n-1));
    // symbolically right-divide
    [q1, r1] := (Ansatz/P);
    [q2, r2] := (Ansatz/Q);
    // now, compare coefficients
    conv := (y) ->  expr(y); //mapcoeffs(expr(y), AA, R);
    E := {(conv(coeff(r1, i))) $ i=0..degree(r1),
          (conv(coeff(r2, i))) $ i=0..degree(r2) };
    // solve the linear equations
    Sol := linsolve(E, AA);
    d := nops(_union(op(map(Sol, indets@rhs))))-1;
    n := n-d;
  until d = 0 end_repeat;
  dom::new(subs(expr(Ansatz), Sol));
end_proc; // lclm

//------------------------------------------------------------------
// _invert -- inversion generally impossible
//------------------------------------------------------------------
_invert := (p) -> FAIL;

//------------------------------------------------------------------
// adjoint -- ...
//------------------------------------------------------------------
adjoint := proc(p : dom) 
local i;
begin
  new(dom, poly([ [ (expr(i[1]) | X=-X-i[2]), i[2] ] $ i in poly2list(op(p,1))], [Shift], R)); 
end_proc;

//==================================================================
// Accessing coefficients and properties
//------------------------------------------------------------------
// lterm -- return the leading term
//------------------------------------------------------------------
lterm := proc(p)
local i;
begin
  dom::new(lterm(rep(p), args(i) $ i=2..args(0)));
end_proc; // lterm

//------------------------------------------------------------------
// lmonomial -- return the leading monomial
//------------------------------------------------------------------
lmonomial := proc(p) 
local i;
begin
  dom::new(lmonomial(rep(p), args(i) $ i=2..args(0)));
end_proc; // lmonomial

//------------------------------------------------------------------
// lcoeff -- return the leading coefficient
//------------------------------------------------------------------
lcoeff := proc(p)
local i;
begin
  lcoeff(rep(p), args(i) $ i=2..args(0));
end_proc; // lcoeff

//------------------------------------------------------------------
// tcoeff -- return the trailing coefficient
//------------------------------------------------------------------
tcoeff := proc(p)
local i;
begin
  tcoeff(rep(p), args(i) $ i=2..args(0));
end_proc; // tcoeff

//------------------------------------------------------------------
// poly - return the corresponding polynomial
//------------------------------------------------------------------
poly := (p) -> rep(p);

//------------------------------------------------------------------
// degree -- return the degree in Shift
//------------------------------------------------------------------
degree := proc(p) 
local i;
begin
  return (degree(rep(p), args(i) $ i=2..args(0)));
end_proc;

order := degree;

//------------------------------------------------------------------
// nterms -- return the number of terms in Shift
//------------------------------------------------------------------
nterms := (p) -> nterms(rep(p));

//------------------------------------------------------------------
// indets -- return the indeterminats
//------------------------------------------------------------------
indets := (p) -> {};

//------------------------------------------------------------------
// iszero -- check for being zero
//------------------------------------------------------------------
iszero := (p) -> iszero(rep(p));

//------------------------------------------------------------------
// coeff -- find a coefficient
//------------------------------------------------------------------
coeff := proc(p : dom, n) 
begin
  if (args(0) = 2) then
    return (coeff(rep(p), Shift, n));
  elif args(0) = 1 then
    return (coeff(rep(p)));
  elif args(0) = 3 and args(2)=Shift then
	return (coeff(rep(p), Shift, args(3)));
  else
    error("Wrong number of arguments");
  end_if;
end_proc; // coeff

//==================================================================
// Output and Conversion
//------------------------------------------------------------------
// print
//------------------------------------------------------------------
print := proc(p : dom) 
local i, l, St;
begin
//  return(op(rep(p),1)); 
  
  St := (x) ->
      _if(x[2]=0,
        x[1],
        _if(x[1]=R::one,
          Shift^x[2],
          hold(_mult)(x[1],Shift^x[2])));

  l := poly2list(rep(p));
  if nterms(rep(p)) = 0 then
    0
  elif nterms(rep(p))<=1 then
    St(l[1]);
  else
    hold(_plus)(St(i)$ i in l);
  end_if;
end_proc;

//------------------------------------------------------------------
// TeX
//------------------------------------------------------------------
TeX := proc(p : dom)
local i;
begin
  St := (x) ->
        _if(x[2]=0,
          generate::TeX(x[1]),
          _if(x[1]=R::one,
            generate::TeX(Shift^x[2]),
            _if(testtype(expr(x[1]), Type::Rational),
              generate::TeX(hold(_mult)(x[1],Shift^x[2])),
              "\\left(".generate::TeX(x[1])."\\right)\\,".generate::TeX(Shift^x[2]))));

  l := poly2list(rep(p));
  if nterms(rep(p)) = 0 then
    "0"
  elif nterms(rep(p))<=1 then
    St(l[1]);
  else
    l := map(l, St);
    l[1]._concat((" + ",  l[i]) $ i=2..nops(l));
  end_if;
end_proc;

//------------------------------------------------------------------

// expr
//------------------------------------------------------------------
expr :=  (p) -> expr(rep(p));

//------------------------------------------------------------------
// mapcoeffs
//------------------------------------------------------------------
mapcoeffs := proc(p:dom) 
local i;
begin
  mapcoeffs(rep(p), args(i) $ i=2..args(0));
  new(dom, %);
end_proc; // mapcoeffs

//------------------------------------------------------------------
// _mod
//------------------------------------------------------------------ 
modp := dom::_mod;
_mod := proc(r : dom, e : DOM_INT) 
  local p;
  begin
    p := mapcoeffs(rep(r), _mod, e);
    return(new(dom, p)); 
  end_proc; // _power

//------------------------------------------------------------------
// polyfy
//------------------------------------------------------------------
polyfy := proc(p : dom) 
local t;
begin
  t := R::lcdenom(coeff(p));
  if t = FAIL then
    t := _mult(map(coeff(p), denom));
  end_if;
  [mapcoeffs(p, _mult, t), R(t)];
end_proc; // polyfy

//------------------------------------------------------------------
// denombound -- returns a polynomial which is a multiplicative 
//                bound for the denominator of any rational solution
//------------------------------------------------------------------
denombound := proc(p : dom)
local co, i, t, B0, B1, a, b, base, bound, f, lc, tc, r, l1, l2;
save T, L, bc;
begin
  delete T,L;
  p := dom::polyfy(p)[1]; // don't care about factor for the moment
  lc := polylib::shift(R::numer(dom::lcoeff(p)), -dom::degree(p));// | X = X-dom::degree(p), [X], F);
  tc := numer(dom::tcoeff(p));
  // base shall become a list of representatives of the occurring shift-
  // equivalence-classes.
  r := dom::baserep(lc, tc);
  base := r[1];
  [B0, B1] := r[3];
  co := table(L=r[2][1][2], T=r[2][2][2]);
  // co[L] and co[T] contain a list for the leading/trailing coeff resp.
  // each element is a list that stands for one shift-equivalence-class
  // determined by the 1st element, the basevector.
  // within these classes every element is a list of the form
  // [shift, multiplicity] corresponding to a factor 
  // basevector(x-shift)^multiplicity.
  // now, go thru the factors again...
  bound := 1;
  B0 := B0;
  B1 := B1;
  for b in base do
    l1 := array(B0..B1, [0 $ B1-B0+1]);
    // going up
    f := co[T][b];
    if f = FAIL then next end_if;
    for i from B0 to B1 do
      t := 0;
      for a in f do
        if i >= a[1] then
          t := t + a[2]; 
        end_if;
      end_for;
      l1[i] := t;
    end_for;
    // going down
    l2 := array(B0..B1, [0 $ B1-B0+1]);
    f := co[L][b];
    if f = FAIL then next end_if;
    for i from B1 downto B0 do
      t := 0;
      for a in f do
        if i <= a[1] then
          t := t + a[2]; 
        end_if;
      end_for;
      l2[i] := t;
    end_for;
    bound := bound * _mult( expr(polylib::shift(b, i))^min(l1[i], l2[i]) $ i=B0..B1 );
  end_for; // shifteq-classes
  return (bound);  
end_proc; // denombound

//------------------------------------------------------------------
// valbound -- returns a table indexed by the factors of the 
//             leading/trailing coefficients. For each base-vector
//             it returns the max. valuation to the left/right
//------------------------------------------------------------------
valbound := proc(p : dom)
local co, B0, B1, b, base, bound, lc, tc, r, zeroiffail;
save T, L;
begin
  delete T,L;
  p := dom::polyfy(p)[1]; // don't care about factor for the moment
  lc := polylib::shift(R::numer(dom::lcoeff(p)), -dom::degree(p));// | X = X-dom::degree(p), [X], F);
  tc := numer(dom::tcoeff(p));
  // base shall become a list of representatives of the occurring shift-
  // equivalence-classes.
  r := dom::baserep(lc, tc);
  base := r[1];
  [B0, B1] := r[3];
  co := table(L=r[2][1][2], T=r[2][2][2]);
  // co[L] and co[T] contain a list for the leading/trailing coeff resp.
  // each element is a list that stands for one shift-equivalence-class
  // determined by the 1st element, the basevector.
  // within these classes every element is a list of the form
  // [shift, multiplicity] corresponding to a factor 
  // basevector(x-shift)^multiplicity.
  // now, go thru the factors again...
  bound := table();
  zeroiffail := (x) -> _if(x=FAIL, 0, _plus(op(map(x, op, 2))));
  for b in base do
    bound[b] := (-zeroiffail(co[L][b]))..zeroiffail(co[T][b]); 
  end_for; // shifteq-classes
  return (bound);  
end_proc; // valbound

//------------------------------------------------------------------
// baserep -- computes the representation of one or more polynomials
//             wrt. a base of non-shift-eq. polys.
// returns [base, L, [B0,B1]]
// where base is the used base, B0 and B1 are the minimal/maximal
// used shifts respectively and L is a List containing a list for
// each argument. 
// example: 
// baserep(SR::baserep(poly(3*(x+1)*(x-2)*(x^2+1)*((x + 3)^2 + 1)), poly(x-29)))
// yields
// [ [x - 2, x^2 + 1], 
//   [ [3, table((x^2 + 1)=[[0, 1], [3, 1]],
//               (x - 2)=[[0, 1], [3, 1]])], 
//     [1, table((x^2 + 1)=FAIL,
//               (x - 2)=[[-27, 1]])]
//   ], 
//   [-27, 3] 
// ]
// (poly() removed for simplicity)
// Hoeij uses an actual zero of the irreducible factors, so his
// 0 corresponds to x here.
//------------------------------------------------------------------
baserep := proc()
local Rep, B0, B1, b, base, cc, cn, co, findshift, i, j, t, t2, t3, w, s;
option remember;
begin
  P := [ args() ];
  if {op(map(P, domtype))} <> {DOM_POLY} then
    error("Wrong type of parameters");
  end_if;
  findshift := proc() option remember; begin polylib::shiftequiv(args()) end_proc;
  P := map(P, factor);
  // base shall become a list of representatives of the occurring shift-
  // equivalence-classes.
  base := _concat(op(map(P, Factored::factors)));
  //base := _concat(op(map(P, (y) -> _if(nops(y) = 1, [1], Factored::factors(y)))));
  for i from 1 to nops(base) do
    if base[i] = FAIL then next end_if;
    for j from i+1 to nops(base) do
      if base[j] = FAIL then next end_if;
      if (s := findshift(base[i], base[j])) <> FAIL then
        // make sure we have only negative shifts
        if s < 0 then
          base[i] := base[j];
        end_if;
        base[j] := FAIL;
      end_if;
    end_for;
  end_for;
  base := select(base, _unequal, FAIL);
  // Rep will contain a list for each argument
  // each element is a list that stands for one shift-equivalence-class
  // determined by the 1st element, the basevector.
  // within these classes every element is a list of the form
  // [shift, multiplicity] corresponding to a factor 
  // basevector(x-shift)^multiplicity.
  B0 := infinity;
  B1 := -infinity;
  Rep := [];
  for w in P do 
    co := [ op(w) ];
    cn := (nops(co)-1)/2;
    cc := co[1];
    t := [ [co[2*i], co[2*i+1]] $ i=1..cn ];
    co := table();
    for b in base do
      t2 := [];
      for j from 1 to nops(t) do
        if t[j] = FAIL then next end_if;
        if FAIL <> (t3 := findshift(b, t[j][1])) then
          t2 := t2.[ [t3, t[j][2]] ];
          B0 := min(B0, t3);
          B1 := max(B1, t3);
          t[j] := FAIL;
        end_if; // same shifteq-class
      end_for; // list of factors
      if nops(t2) > 0 then
        co[b] := t2;
      else
        co[b] := FAIL;
      end_if;
    end_for; // list of shifteq-classes
    Rep := Rep.[ [cc, co] ];
  end_for; // args
  return([base, Rep, [B0, B1]]);
end_proc;

//------------------------------------------------------------------
// symprod(p,q) -- compute the symmetric product of p and q 
//                  where q
//                  has order 1.
//                  cf. CluHoe05, p5
//------------------------------------------------------------------
symprod := proc(p : dom, q : dom) 
local r, P, i, n, B, j;
save a,b,c;
begin
  if dom::degree(q) = 1 then 
    n := degree(p);
    r := -coeff(q,0)/coeff(q,1);
    B := array(0..n, [(coeff(p,i) * _mult((expr(r) | X=X+j) $ j=i..n-1) ) $ i=0..n-1, coeff(p,n)]);
    P := dom::new(_plus( B[i]*Shift^i $ i=0..n ) );
    return(P);
  else
    error("not yet implemented");
    /*delete a,b,c;
    f := op(p,1);
    g := op(q,1);
    m := degree(f);
    n := degree(g);
    s1 := -(f(a)-lmonomial(f(a),a))/lcoeff(f);
    s2 := -(g(b)-lmonomial(g(b),b))/lcoeff(g);
    F := poly(_plus(c.i*(a*b)^i $ i=0..m*n), [a,b]  );
    while degree(F,a) >= m or degree(F,b) >= n do
      F2 := poly(0, [a,b]);
      for t in poly2list(F) do
        [da, db] := t[2];
        poly(t[1] * _if(da >= m, a^(da-m)*s1, a^da) * _if(db >= n, b^(db-n)*s2, b^db), [a,b]);
        F2 := F2 + %;
      end_for;
      F := F2;
    end_while;  
    L := { t[1] = 0 $ t in poly2list(F) };
    G := subs(_plus(c.i*X^i $ i=0..m*n), op(linsolve(L)));
    G := subs(G, { i=1 $ i in (indets(G) minus {X}) });
    poly(G,[x]);*/
  end_if;
end_proc; // symprod

//------------------------------------------------------------------
// polysols -- returns the polynomial solutions of the SO
//------------------------------------------------------------------

polyupperbound := proc(L) 
local i, j, N, n, a, b;
begin
  ord := dom::order(L);
  a := array(0..ord, [numer(coeff(L, i)) $ i=0..ord]);
  d := map(a, degree, X):
  m:=max(d[i] $ i=0..ord)+1;
  alpha := genident("alpha");
  repeat
    m:=m-1;
    /* compute the coefficient of n^(alpha+m) if deg(u(n),n)=alpha */
    s:=0;
    for i from 0 to ord do
      /* [n^(alpha+m)] l[i]*(n+i)^alpha */
      for j from max(0,m) to d[i] do
        s:=s+coeff(a[i],j)*binomial(alpha,j-m)*i^(j-m);
      end_for
    end_for;
    s := expand(s);
  until s <> 0 end_repeat;
  t := -1;
  if has(s, alpha) then
    s := polylib::primpart(s, [alpha]);
    /* we cannot correctly handle parametric or algebraic coefficients */
    if testtype(s, Type::PolyExpr([alpha], Type::Rational)) then
      s := solvelib::iroots(s);

      /* t := maximal degree of a polynomial solution of the homogeneous
         equation */
      t := max(-1, op(s));
    end_if;
  end_if;

  /* t := maximal degree of a polynomial solution of the inhomogenenous
     equation */
  //t := max(t, max(d[i] $ i=0..ord) - m);
  t;
end_proc;

polysols := proc(L:dom, isPoly=FALSE, bound=FAIL)
local i,n,N,pol,eq,sol,Y,j;
save a;
begin
  delete a;
  if not isPoly then
    L := dom::polyfy(L)[1];
  end_if;
  Y := indets(expr(L));
  n := dom::order(L);
  if bound = FAIL then
    N := dom::polyupperbound(L);
  else 
    N := bound;
  end_if;
  //print(N);
  if N<0 then 
     return({});
  end_if;
  pol := _plus(`#`.i*X^i $ i=0..N);
  eq := [ coeff(collect(L(pol), X), [X]) ];
  //warning("halt!");
  sol := numeric::linsolve(eq, [`#`.j $ j=0..N], Symbolic);
  sol := normal(subs(pol, sol));
  select({ normal(subs(sol, [`#`.j=1 $ j=0..N])) }, (A) -> not iszero(A));
end_proc; // polysols


polysols_gsss := proc(p : dom)
local Sol, r, ind;
save Y;
begin
  delete Y;
  r := rec(dom::toRec(p, Y), Y(X));
  Sol := solve(rec(dom::toRec(p, Y), Y(X)));
  if has(Sol, solve) then
    return({});
  end_if;
  ind := indets(Sol) minus {X, Shift};
  eval(subs(Sol, map(ind, _equal, 1)));
  if Sol={0} then Sol := {} end_if;
  return (Sol);
end_proc;

//------------------------------------------------------------------
// ratsols -- returns the rational solutions of the SO
//------------------------------------------------------------------
ratsols := proc(p : dom)
local D, B, n, P, i, Y, pSol, rSol, t;
begin
  Y := indets(expr(p)) minus {X, Shift};
  if nops(Y) > 0 then
    t := dom::coeff(p);
    for i in Y do
      i^max(map(t, (x) -> degree(denom(normal(expr(x))),[op(Y)],i)));
      p := dom(%)*p;
    end_for;
  end_if;
  D := dom::denombound(p); // Denominator
  n := degree(p);
  B := array(0..n, [ coeff(p, i) / (D | X=X+i) $ i=0..n ]);
  P := dom::new(_plus( expr(B[i])*Shift^i $ i=0..n ));
  P := dom::polyfy(P)[1];
  Y := genident("Y");
  pSol := dom::polysols(P);
  rSol := [FAIL $ i=1..nops(pSol)];
  for i from 1 to nops(pSol) do
    P := pSol[i];
    Y := indets(P);
    rSol[i] := P/D;
  end_for;
  return({op(rSol)});
end_proc; // ratsols

getfactors := proc(a)
local F, i, lc, n;
begin
  lc := lcoeff(a);
  F := factor(multcoeffs(a, 1/lc));
  n := (nops(F)-1)/2;
  F := { (op(F,2*i) $ op(F, 2*i+1)) $ i=1..n };
  return([lc, [op(map(combinat::subsets(F), _mult@op))]]);
end_proc;

petkovsek := proc(L : dom, Method=FAIL)
local Coo, Lc, Paoo, Soo, a, b, ca, cb, f, i, j, l, m, n, p, paoo, polrec, polsol, s, sol, u, x, z, zsol;
begin
  z := genident("sdd");
  J := 0;
  JJ := 0;
  JJJ := 0;
  L := dom::polyfy(L)[1];
  paoo := dom::localTypeAtInifinity(L);
  Paoo := table(op(x,2) = {} $ x in paoo);
  for x in paoo do
    if not iszero(x) then
      Paoo[op(x,2)] := Paoo[op(x,2)] union { normal(op(x,1)) }
    end_if
  end_for;
  Soo := { op(map(paoo, op, 2)) };
  n := dom::order(L);
  Lc := [expr(coeff(L,i)) $ i=0..n];
  a:=poly(Lc[1], [X]);
  b:=poly(subs(Lc[n+1], X=X-n+1),[X]);
  [ca,a]:=(dom::getfactors(a,[X])); // generate a list of all combinations of factors of lowest coefficient polynomial
  [cb,b]:=(dom::getfactors(b,[X])); // generate a list of all combinations of factors of highest coefficient polynomial
  [a,b] := [map(a, expr), map(b, expr)];
  sol := {};
//  print(Unquoted, "Insgesamt: ".expr2text(nops(a)*nops(b))." Möglichkeiten");
  for i from 1 to nops(a) do
    for j from 1 to nops(b) do
      J := J +1;
      s := -(degree(a[i],X)-degree(b[j],X));
      if Method <> Traditional then
        if not s in Soo then 
          next;
        end_if;
      end_if;
      JJ := JJ + 1;
      f := [(Lc[l+1]*_mult((a[i] | X=X+m) $ m=0..l-1)*_mult((b[j] | X=X+m) $ m=l..n-1)) $ l=0..n]; // coefficient polynomials for recurrence of polynomial c(x)
      if Method=Traditional then
        m := max(op(map(f, degree, X)));
        u := [ coeff(f[l+1],X,m) $ l=0..n];
        u := poly([ [coeff(f[l+1],X,m), l] $ l=0..n], [X], Expr); // $ m=0..n)
        zsol := solve(u(z)=0, z);
      else
        zsol := [normal(x*lcoeff(b[j])/lcoeff(a[i])) $ x in Paoo[s]];
      end_if;
      for p in zsol do
        //if not iszero(u(p)) then print("sdsd"); next; end;
        JJJ := JJJ + 1;
        polrec:=new(dom, poly([[expand(_plus(p^s*expr(f[s+1]))), s] $ s=0..n], [Shift])); 
        //print(polrec);
        polsol:=dom::polysols(polrec, TRUE); 
        if nops(polsol)=0 then polsol := 0 else polsol := op(polsol, 1) end_if;
        if has(polsol,X) or polsol<>0 then
          //print (normal(lcoeff(a[i])/lcoeff(b[j])*p), Coo, p);
          sol:=sol union {p*a[i]/b[j]*(polsol | X=X+1)/polsol}; // create normal form z*a(x)/b(x)*c(q*x)/c(x)
        end_if;
      end_for;
    end_for;
  end_for;
  map(sol, (x)->dom::new(Shift-x));
end_proc;




myIsInteger := proc(e, epsilon=0.01)
local i;
begin
  for i in float(e) do
    if domtype(i)<>DOM_FLOAT and domtype(i)<>DOM_COMPLEX then
      return(TRUE);
      print(e);
      print(float(e));
      print(i);
    end_if;
    if abs(Im(i)) < 0.1 then
      if abs(i-round(i)) < 99 then // epsilon then
        return(TRUE)
      end_if;
    end_if;
  end_for;
  return(FALSE);
end_proc;

//------------------------------------------------------------------
// factorHoeij -- returns right-factors of order 1
//                This is van Hoeij's algorithm
//------------------------------------------------------------------
factorHoeij := proc(L : dom, Method=FAIL)
local i, L2, RF, SS, c, cand, d, l, pai, r, s, sings, sols, t, 
  vb, S, va, J, JJ, FE, RS, r1, r2, R2;
begin
  // we assume that there are no rational or polynomial solutions
  vb := dom::valbound(L);
  if FALSE and 100 < _mult(op(rhs(i),2)-op(rhs(i),1) $ i in vb) then
    // evil heuristic: The problem-points don't change if we compute the lclm
    // with an Shift-someinteger, but there is a hope that the candidate-types 
    // become fewer.
    // there are so many possibilities that the lclm-trick is worth the efford
    L2 := dom::lclm(L, dom::new(Shift-3));
    va := dom::valbound(L2);
    t := cgenerator({op(vb)}, {op(va)});
    vb := table();
    while FAIL <> (i := t()) do
      if FAIL <> polylib::shiftequiv(lhs(i[1]),lhs(i[2])) then
        r1 := rhs(i[1]); r2 := rhs(i[2]);
        r := max(op(r1,1), op(r1,1))..min(op(r1,2), op(r1,2));
        if op(r,1) <= op(r,2) then
          vb[lhs(i[1])] := r;
        end_if;
      end_if;
    end_while;
  end_if;
  pai := dom::localTypeAtInifinity(L);
  sings := [lhs(t) $ t in vb];
  sols := {};
  J := 0; JJ := 0;
  J := nops(pai) * _mult(1+op(rhs(t),2)-op(rhs(t),1) $ t in vb);
  userinfo(2, expr2text(J)." Versuche insgesamt.");
  
  for l in pai do
    [c,s,d] := l;
    cand := dom::cgenerator({i$i=rhs(t)} $ t in vb);
    while FAIL <> (t := cand()) do
      J := J-1;
      if J mod 10 = 1 then
        userinfo(3, "Noch ".J." Kombinationen.");
      end_if;
      if s+_plus(degree(sings[i])*t[i] $i=1..nops(t)) <> 0 then next end_if; // 1st fuchs-relation
      if FALSE and dom::myIsInteger(d+_plus(coeff(sings[i], degree(sings[i])-1) $ i=1..nops(t)), 1) then 
        next;
      end_if; // 2nd fuchs-relation
        JJ := JJ+1;
        userinfo(3, "Schon ".JJ." Kombinationen getestet.");
      // now, we know that both fuchs-relations hold and we construct an Operator of 
      // appropriate type
      if (domtype(c) in {DOM_INT, DOM_RAT, F}) or F=Dom::ExpressionField() then 
        FE := F;
        R2 := R;
        RS := dom;
      else
        FE := domtype(c); // we possibly need to deal with field-extensions
        R2 := Dom::UnivariateRationalFunctions(X, FE);
        RS := Dom::ShiftOperators(X, Shift, FE);
      end_if;
      r := R2(1/c) * R2(1/expr(_mult(sings[i]^t[i] $ i=1..nops(t))));
      RF := RS(Shift-r);
      L2 := RS::symprod(RS(L), RF);
//      print(c, L2);
      SS := RS::ratsols(L2);
      if nops(SS) > 0 then
        for S in SS do 
//          print ("----", S);
          if nops((va := indets(SS) minus {X})) > 0 then
            S := subs(S, i=1 $ i in va);
          end_if;
          if iszero(S) then
            next;
          else
            sols := sols union { RS(Shift - subs(S, X=X+1)/S/expr(r)) };
            if Method=PrincipalValue then
              return(sols);
            end_if;
          end_if;
        end_for
      end_if;
    end_while;
  end_for;
  userinfo(3, "Gesamtkombinationen probiert: ".JJ);
  return(sols);
end_proc;

cgenerator := proc()
local word, word_int, A, i, first_time, s;
option escape;
begin
    if args(0)=0 then return(FAIL); end_if;
	A:=[args()]:
    for i from 1 to nops(A) do
        if domtype(A[i])=DOM_INT then
            A[i]:={$1..A[i]};
    elif testargs() and not (testtype(A[i], DOM_LIST) or testtype(A[i], DOM_SET)) then
	error("Expecting an integer, list, or set; got: ".expr2text(A[i]));
        end_if;
        if nops(A[i])=0 then
            return(FAIL);
        end_if;
    end_for;
    word_int := [1 $ args(0)];
    word     := [op(s,1) $ s in A];
    first_time:=TRUE;
    // Invariants:
    //  - word_int is a sequence of k integers with
    //		1 <= word_int[i] <= nops(A[i])
    //  - word is the corresponding element of A[1] x ... x A[k]
    //  - first_time is true if and only if it's the first run
    return(proc()
               local index;
           begin
               if first_time then
                   first_time:=FALSE;
                   return(word);
               end_if;
               for index from nops(A) downto 1 do
                   if word_int[index] < nops(A[index]) then
                       word_int[index] := word_int[index]+1;
                       word[index] := op(A[index], word_int[index]);
                       return(word);
                   else 
                       word_int[index] := 1;
                       word[index] := op(A[index],1);
                   end_if;
               end_for;
               return(FAIL);
           end_proc)
end_proc; // cgenerator


//------------------------------------------------------------------
// factorOrder2 -- 
//------------------------------------------------------------------
rightFactorOrder := proc(L : dom, ord)
local L2, P1, SS, Shift_on_b, W, b, b2, bT, bb, bn, bt, bv, c, cv, 
  eP, eq, fl, i, j, l, leftfactors, m, mop, mopo, rp, s, sgn, sol, 
  vrs, vv, yv, invShift_on_b, vlen, mi, ma, Sol, bbx, einsinv, x;
save S, rho, t;
begin
  delete S, rho, t;
  userinfo(1, "called rightFactorOrder");

  m := dom::order(L);

  // Handle some special cases
  if m = ord then
    return({L});
  elif m < ord then
    warning ("Operator of order ".m." won't have factors of order ".ord.".");
    return({});
  elif ord = 1 then
    return(dom::factorHoeij(L));
  elif ord > m/2 then
    return(map(dom::rightFactorOrder(dom::adjoint(L), m-ord), dom::adjoint));
  end_if;

  // initialize
  mopo := binomial(m,ord);
  L := mapcoeffs(L, _divide, lcoeff(L));

  c := array(0..m, [expr(coeff(L,i)) $ i = 0..m]);
  s := array(0..m-1, [`#shift^`.i $i=0..m-1]);
  eP := map(combinat::choose({i $ i=0..m-1}, m-ord), (x)->sort([op(x)]));
  mopo := nops(eP);
  bT := table((eP[i+1] = `#b`.i) $ i=0..mopo-1);
  //bT := table((i = `#`._concat("s".expr2text(i[j])."^" $ j=1..m-3).i[m-2]) $ i in eP );
  bb := array(0..mopo-1, [rhs(b) $ b in bT]);
  
  //
  // Preparations to handle arbitrary order
  //
  // Compute the action of Shift on the base-vectors of the external product
  //
  Shift_on_b := [ X = X+1 ];
  for b in bT do
    bn := rhs(b);
    bv := lhs(b);
    b2 := map(bv, _plus, 1);
    if type(bT[b2]) <> "_index" then
      // lucky case, shift maps b to the element b2
      Shift_on_b := Shift_on_b . [ bn = bT[b2] ];
      next;
    end_if;
    // complicated case, shift maps b to a linear combination
    bt := 0;
    for l in [b2[1..-2].[j] $ j in select([i $ i=0..m-1], (x)->not(has(b2[1..-2], x)))] do
      sgn := (-1)^(1+nops(l)-contains(sort(l),l[-1]));
      bt := bt + sgn * c[l[-1]]*bT[sort(l)];
    end_for;
    Shift_on_b := Shift_on_b . [ bn = bt ];
  end_for;
  //
  // What does Shift^-1 do on an Vector?
  //
  invShift_on_b := [ X = X-1 ];
  einsinv := poly(_plus(subs(-1/c[0], X=X-1)*S^(m-1) , -subs(c[i+1]/c[0], X=X-1)*S^i $ i=0..m-2), [S], Expr);
  for b in bT do
    bn := rhs(b);
    bv := lhs(b);
    b2 := map(bv, _plus, -1);
    if type(bT[b2]) <> "_index" then
      // lucky case, shift maps b to the element b2
      invShift_on_b := invShift_on_b . [ bn = bT[b2] ];
      next;
    end_if;
    // complicated case, shift maps b to a linear combination
    bt := 0;
    for l in [[j].b2[2..-1] $ j in select([i $ i=0..m-1], (x)->not(has(b2[2..-1], x)))] do
      t := sort(l);
      sgn := (-1)^(1+contains(t,l[1]));
      bt := bt + sgn * coeff(einsinv, l[1])*bT[t];
    end_for;
    invShift_on_b := invShift_on_b . [ bn = bt ];
  end_for;
  //return([Shift_on_b, invShift_on_b]);
  //
  // find how to get the coefficients of an 2nd-order-factor from the eigenvector
  //
  // assume the ord-th-order factor is 
  // Shift^ord + rho[ord-1]*Shift^ord[-1] + ... + rho[0]
  // then, the eigenvector has the form
  // (Shift^ord + rho[ord-1]*Shift^(ord-1) + ... + rho[0]) 
  //   ^ (Shift^(ord+1) + rho[ord-1]*Shift^(ord) + ... + Shift*rho[0])
  //     ^ ... ^ (Shift^(m-1) + rho[ord-1]*Shift^(m-2) + ... + Shift^(m-ord)*rho[0]) 
  SS := [ S[i] $ i=0..m-1 ];
  W := poly(_mult( _plus(_if(j=ord,1,rho[j])*S[j+i] $ j=0..ord) $ i=0..m-(ord+1) ), SS);
  W := select(poly2list(W), (x) -> _lazy_and(max(op(op(x,2)))<2, type(op(x,1)) in {"_index", DOM_INT} ));
  if nops(W) <> ord+1 then error("This ought to work!") end_if;
  W := table( (t[1]=bT[select([i $ i=0..m-1], (i)->(t[2][i+1]=1))]) $ t in W );
  // end preparations for arbitrary order
  
  userinfo(1, "initialization done.");
  
  // Compute a cyclic vector and the corresponding minimal operator
//  MOP := [0,0,0,0];
  S := {};
  bbx := [op(bb), X];
  i := 0;
	Sol := {};
//  for J from 1 to 4 do
	repeat
	  userinfo(1, "trying to find a cyclic vector ".expr2text(i));
	  // first, choose a vector that shall be cyclic. 
	  repeat
	    i := i+1;
  	  if i<mopo then
  	    cv := bb[i-1];
  	  elif i < mopo+mopo^2 then
  		  cv := bb[i-mopo mod mopo]+X*bb[i-mopo div mopo];
  	  else
  	    t := i-(mopo+mopo^2)+1;
  	    rp := (x) -> expr(polylib::randpoly([X], Degree=t+1, Terms=t, Coeffs=random(-t..t), Monic));
  		  cv := _plus(bb[j]*rp() $ j=0..mopo);
  	  end_if;
  	until not cv in S end_repeat;
  	cv := poly(cv, bbx);
    if cv in S then next end_if;
  	S := S union {cv};
	  // compute the iterated Shifts on the cyclic vectors
	  vv := array(-mopo..mopo, [0 $ mopo, cv, 0 $ mopo]);
	  vlen := array(-mopo..mopo, [0 $ mopo, length(cv), 0 $ mopo]);
	  for j from 1 to mopo do
	    vv[j] := poly(expand(eval(subs(expr(vv[j-1]), Shift_on_b))), [op(bb)]);
	    if vv[j] in S then next end_if;
	    S := S union {vv[j]};
	    vlen[j] := length(vv[j]);
	  end_for;
	  for j from -1 downto -mopo do
	    vv[j] := poly(expand(eval(subs(expr(vv[j+1]), invShift_on_b))), [op(bb)]);
	    if vv[j] in S then next end_if;
	    S := S union {vv[j]};	    
	    vlen[j] := length(vv[j]);
	    if vlen[j] > vlen[j+mopo+1] then break; end_if;
	  end_for;
	  mi := 0; ma := 0;
	  while ma-mi < mopo do
	    if vlen[ma+1] <= vlen[mi-1] then
	      ma := ma + 1;
	    else
	      mi := mi - 1;
	    end_if;
	  end_while;
	  vv := array(0..mopo, [vv[i] $ i=0..mopo]);
	  //vv := array(0..mopo, [vv[i] $ i=mi..ma]);
	  cv := vv[0];
	  // set up the equations and solve the linear systems
	  vv := map(vv, op, 1);
	  yv := array(0..mopo, [genident("Y") $ i=0..mopo]);
	  eq := [ coeff(_plus(yv[j]*vv[j] $ j=0..mopo), [op(bb)]) ];
	  //sol := solve(eq, [op(yv)], IgnoreSpecialCases);
	  sol := liinalg::linsolve(eq, [op(yv)]);
	  if nops(sol) > 1 then next end;
	  mop := subs(_plus(yv[j]*Shift^j $ j=0..mopo), op(sol,1));
	  vrs := indets(sol) minus { op(yv), X, Shift };
	  userinfo(1, "Linear System leads to ".expr2text(nops(vrs))." free parameter(s).");
	  // set up the equations and solve the linear systems
	until nops(vrs)=0 and coeff(mop,Shift,0) <> 0 end_repeat;
	mop := dom::new(mop); //subs(mop, [vrs[1]=1]));
	// now, cv is a cyclic vector and 
	// mop the corresponding minimal operator of order 6
  userinfo(1, "Found cyclic vector and computed minimal operator.");
  userinfo(1, mop);

  // find 1st-order left(!)factors
  //
  // try factoring the reduced minimal operator
  // 
  leftfactors := dom::factorHoeij(dom::adjoint(mop), PrincipalValue);
  userinfo(1, "Found left-factors: ".expr2text(leftfactors));
  if nops(leftfactors)=0 then return ({}); end_if;

  
  for fl in leftfactors do
    // leftdivide by this factor
    P := dom::leftdivide(mop, dom::adjoint(fl));
    // check it's all right
    if not iszero(P[2]) then error("can-not-happen-error"); end_if;
    P := P[1];
  
    // Now, we know that P(cv) is an eigenvector of Shift in the 
    // 2nd external product.
    P1 := poly((_plus(expr(coeff(P,i)) * vv[i] $ i=0..mopo-1)), [op(bb)]);
//    beta := map(array(0..mopo-1, [coeff(P1, b.i) $ i=0..mopo-1]), poly, bb);
//    if 0=degree(beta[0]*beta[5]+beta[2]*beta[3]-beta[1]*beta[4]) then
    // cv is an eigenvector that corresponds to a 2nd-order factor.
    t := expr(coeff(P1, W[1], 1));
    L2 := dom::new(Shift^ord + _plus(expr(coeff(P1, W[rho[i]], 1))/t*Shift^i $ i=0..ord-1));
    if iszero(dom::rightdivide(L,L2)[2]) then
      Sol := Sol union { L2 };
    end_if;
  end_for;
  return (Sol);

end_proc; // factorOrder2test



//------------------------------------------------------------------
// factor -- completely factor operators of order 4, try the best 
//           for lower order
//------------------------------------------------------------------
factor := proc(L : dom)
local lhf, order, rhf;
begin
  // there's nothing to do.
  if order(L) = 0 or order(L) = 1 then
    return([L])
  end_if;
  
  // right-factors of order 1
  rhf := dom::factorOrder1(L);
  if nops(rhf) > 0 then
    rhf := op(rhf, 1);
    lhf := dom::rightdivide(L, rhf);
    if not iszero(lhf[2]) then error("Found wrong factor"); end_if; 
    return(dom::factor(lhf[1]) . [rhf]);
  end_if;

  // left-factors of order 1
  lhf := dom::factorOrder1(dom::adjoint(L));
  if nops(lhf) > 0 then
    lhf := dom::adjoint(op(lhf, 1));
    rhf := dom::leftdivide(L, lhf);
    if not iszero(rhf[2]) then error("Found wrong factor"); end_if; 
    return([lhf] . dom::factor(rhf[1]));
  end_if;
  
  // order 3 is irreducible if there are no order 1 factors
  if dom::order(L) = 3 then
    return([L]);
  end_if;

  // we don't know how to deal with higher order
  if dom::order(L) > 4 then
    warning("Result is possibly reducible, we just don't know.");
    return([L]);
  end_if;
  
  // at this point, L has order 4 and no order 1 left- or rightfactors
  
end_proc; // factor

//------------------------------------------------------------------
// polygon -- compute the newton-polygon 
//             returns the integer slopes of the Newton-Polygon and
//             the corresponding Newton-Polynomials
//             cf. CluHoe05, p13
//------------------------------------------------------------------
polygon := proc(p : dom)
local a, i, m, n, s, v, Res;
begin
  n := dom::degree(p);
  p := dom::polyfy(p)[1];
  if n=0 then return({}) end_if;
  a := array(0..n, [R::numer(coeff(p, i)) $ i=0..n ]);
  v := array(0..n, [_if(iszero(a[i]), infinity, -degree(a[i])) $ i=0..n]);
  s := max( (v[n]-v[i])/(n-i) $ i=0..n-1 );
  m := 0;
  while (v[n]-v[m])/(n-m) <> s do m := m+1 end_while;
  Res := {};
  if TRUE = is(s, Type::Integer) then
    P := poly(_plus( coeff(a[n-i], s*i-v[n])*X^(n-m-i) $ i=0..n-m ),[X]);
    Res := {[s,P]};
  end_if;
  return(Res union dom::polygon(dom::new(_plus(a[i]*Shift^i $ i=0..m))));
end_proc; // polygon

//------------------------------------------------------------------
// localTypeAtInifinity -- 
//                         cf. CluHoe05, p14
//------------------------------------------------------------------
localTypeAtInifinity := proc(p : dom) 
local A, L, Lcv, Polygon, SR, c, f, s, t, v, D, FE, IE, Lcvd, r, factors, l, AEF, alpha, R2, tmp, x;
begin
  A := {};
  Polygon := dom::polygon(p);
  for t in Polygon do
    [s, P] := t;
    v := -s;
    P := factor(P);
    for f in Factored::factors(P) do
      C := solve(expr(f)=0, X, IgnoreSpecialCases);
/*      if (degree(f) = 1) then 
        // don't need to take an extension
        FE := F;
        SR := dom;
        R2 := R;
        L := p;
        C := {-f(0)/lcoeff(f)}; // root of the linear factor
      else 
      end_if;
        // algebraic extension necessary
        if F <> Dom::ExpressionField() then
          c := genident("c");
          f := ( expr(f) | X=c );
          FE := Dom::AlgebraicExtension(F, f, c);
          SR := Dom::ShiftOperators(X, Shift, FE);
          R2 := Dom::UnivariateRationalFunctions(X, FE);
          c := FE(c);
          L := SR(expr(p));
        else
          FE := F;
          SR := dom;
          R2 := R;
          L := p;
          c := RootOf(f);
        end_if;
      end_if; */
      // Now, L 
      FE := F; //Dom::ExpressionField();
      SR := dom;
      R2 := R; //Dom::UnivariateRationalFunctions(X, FE);
      L := p;
      for c in C do
        tmp := R2(X^v) * (FE(1/c));
        tmp := SR::symprod(L, SR(Shift - tmp));
        Lcv := SR::polyfy(tmp)[1];
        D := genident("delta");
        Lcvd := poly(subs(expr(Lcv), Shift = D+1), [X, D], FE);
        IE := SR::indicialEquation(Lcvd);
        factors := dom::baserep(IE)[2][1][2];
        for f in factors do 
          // we only want to know one actual root, shift doesn't matter
          l := lhs(f);
/*          if degree(l) = 1 then
            r := coeff(l, 0) / coeff(l,1);
          else
            // we need an algebraic extension to represent a zero
            alpha := genident("alpha");
            poly(expr(l) | op(l,2)[1] = alpha, [alpha], FE);
            AEF := Dom::AlgebraicExtension(FE, %);
            r := AEF(alpha);
          end_if; */
          for r in solve(expr(l)=0, op(op(l,2)), IgnoreSpecialCases) do
            A := A union {[c,v,r]};
          end_for;
        end_for;
      end_for;
    end_for; // f in Factored::factors(P)
  end_for;
  return(A);
end_proc; // localTypeAtInifinity

//------------------------------------------------------------------
// indicialEquation -- 
//                      cf. CluHoe05, p14f
//------------------------------------------------------------------
indicialEquation := proc(L) 
local FE, x, d, m, i, a, j, Subs, p, l;
begin
  // L is poly(..., [X, D], FE)
  FE := op(L,3);
  [x, d] := op(L,2);
  a := poly2list(L);
  // a is of the form [ [cij, [i,j]], ...  ]
  m := max(op(map(a, (i) -> i[2][1]-i[2][2])));
  a := select(a, (i) -> m = i[2][1]-i[2][2]);
  d := genident("d");
  Subs := [ genident("a") = i[1] $ i in a];
  p := poly(0, [d]);
  for i from 1 to nops(a) do
    p := p + poly( Subs[i][1] * _mult( (d-j) $ j=0..a[i][2][2]-1 ), [d])    
  end_for;
  l := poly2list(p);
  l := subs(l, Subs);
  p := poly(l, [d], FE)
//  p := poly(p, [d], FE);
end_proc; // indicialEquation

//------------------------------------------------------------------
// toRec -- returns the corresponding expressen for the  rec
//------------------------------------------------------------------
toRec := proc(p : dom, Y)
local i;
begin
  _plus( (expr(coeff(p, i)) * Y(X+i)) $ i = 0..degree(p) )
end_proc; // toRec

//------------------------------------------------------------------
// func_call -- evaluate p(a)
//------------------------------------------------------------------
func_call := proc(p, f, y) 
local m,l;
begin
  f := context(f);
  if args(0) = 3 then
    y := context(y);
  end_if;
  if args(0) < 2 then 
    error("two arguments expected");
  elif args(0) = 2 then
    y := X;
  end_if;
  if domtype(p) <> dom then
    error("Wrong kinds of arguments");
  end_if;
  if (domtype(f) in {DOM_EXPR, DOM_IDENT}) and (has(f,y)) then
    //f := fp::unapply(f, y);
    13
  end_if;
  l := poly2list(extop(p, 1));
  expr(_plus( (m[1] * subs(f, X=X+m[2])) $ m in l));
end_proc; // func_call

// for speed, faster binomials
mybinom := proc(n,k) local i; option remember; begin 
//  _if(k>(n-k), 
//    _mult(i $ i=k+1..n)/_mult(i$i=2..n-k), 
//    _mult(i $ i=n-k+1..n)/_mult(i$i=2..k)) 
  _if(k<=(n-k), 
    _mult(i $ i=k+1..n)/_mult(i$i=2..n-k), 
    dom::mybinom(n, n-k)); 
end_proc;

//==================================================================
//* Domain Constructor
//------------------------------------------------------------------
begin // domain constructor,
  if domtype(X) <> DOM_IDENT or domtype(Shift) <> DOM_IDENT then
    return("First two parameters must be identifiers");
  end_if;
  R := Dom::UnivariateRationalFunctions(x, F);
end:

//==================================================================
// End of file ShiftOp.mu
//==================================================================
unalias (rep):

