//==================================================================
// Dom::qShiftOperators - q-Shift Operators
//------------------------------------------------------------------
// blah...
//------------------------------------------------------------------

// for convenience
alias (rep(p) = extop(p,1)):
                      
domain Dom::qShiftOperators(q=q, X=x, Shift=Eq)

local
    F,
    R,
    P; // Polynomial Ring for internal rep.
       
inherits Dom::BaseDomain; //TruncatedPowerSeries();//[x, xdx], F);
category Cat::Ring;

Indet := X;
characteristic := 0;

//make_slot := proc(d,s) begin print(Unquoted, "Slot requested: ".expr2text(s)); FAIL; end_proc;

whoami := () -> "Dom::qShiftOperators";

//==================================================================
// Creating new Elements
//------------------------------------------------------------------
// new
//------------------------------------------------------------------
new := proc(p) 
local pp;
save f; 
begin
  delete f; 
  if p = FAIL then return(FAIL) end_if;
  if domtype(p) = dom then return(p) end;
  
  pp := collect(expr(p), Shift, normal);
  pp := poly(pp, [Shift], Expr);
  if pp = FAIL then return(FAIL) end;
  if not _lazy_and(op(map([coeff(pp)], c -> testtype(c, Type::RatExpr(X)) and testtype(c, Type::RatExpr(q)))))
    then print("shouldn't convert to qShiftOp.") 
  end;
  new(dom, mapcoeffs(pp, collect, X));
end_proc; // new

//------------------------------------------------------------------
// random
//------------------------------------------------------------------
random := proc()
begin
  FAIL;
  //dom::new(X^mods(random(), 7)*polylib::randpoly([XDX], Monic))
end; // random

//------------------------------------------------------------------
// coerce
//------------------------------------------------------------------
coerce := dom::new;

// Dunno
convert := () -> FAIL;

//==================================================================
// Constants
//------------------------------------------------------------------
one := dom::new(1); 
zero := dom::new(0); 
characteristic := 0;

//==================================================================
// Arithmetics
//------------------------------------------------------------------
// _plus -- Add two or more EuDOs
//------------------------------------------------------------------
_plus := proc()
local a, t, f1, f2, A;
begin
  A := [ _if(domtype(a) = dom, a, _if((t := dom::new(a)) <> FAIL, t, a)) $ a in args()];
  A := split(A, (a) -> domtype(a)=dom);
  f1 := dom::_plus2(op(A[1]));
  if (nops(A[2])+nops(A[3])>0) then
    f2 := _plus(op(A[2].A[3]));
    return(hold(_plus)(f1,f2));
  else
    return(f1);
  end_if;
end_proc; // _plus

_plus2 := proc(p1, p2) local i,t; 
begin
  if args(0) = 1 then
    return(p1)
  end_if;
  if (args(0) > 2) then
    p2 := dom::_plus2(args(i) $ i=2..args(0));
  end_if;
  t := mapcoeffs(rep(p1)+rep(p2), normal);
  return(dom::new(t));
end_proc; // _plus2

//------------------------------------------------------------------
// _negate -- get -L from L
//------------------------------------------------------------------
_negate := (p) -> dom::new(multcoeffs(rep(p),-1));

//------------------------------------------------------------------
// _mult --  Multiply two or more Shift Operators
//------------------------------------------------------------------
_mult := proc()
local a, t, f1, f2, A;
begin
  A := [ _if(domtype(a) = dom, a, _if((t := dom::new(a)) <> FAIL, t, a)) $ a in args()];
  A := split(A, (a) -> domtype(a)=dom);
  f1 := dom::_mult2(op(A[1]));
  if (nops(A[2])+nops(A[3])>0) then
    f2 := _mult(op(A[2].A[3]));
    return(hold(_mult)(f1,f2));
  else
    return(f1);
  end_if;
end_proc; // _mult

_mult2 := proc(p1:dom, p2:dom) 
  local r, l1, l2, e1, c1, e2, c2, t, i, j;
begin
  if args(0) = 0 then
    return(dom::one);
  elif args(0) = 1 then
    return(p1);
  end_if;
  if args(0) > 2 then
    // though not commutative, it's associative
    p2 := _mult(args(i) $ i=2..args(0));
  end_if; 
  r := table();
  l1 := poly2list(rep(p1));
  l2 := poly2list(rep(p2));
  for i in l1 do
    e1 := i[2]; // exponent of S in p
    c1 := i[1]; // coefficient of S^e1 in p
    for j in l2 do
      e2 := j[2]; // exponent of S in q
      c2 := j[1]; // coefficient of S^e2 in q
      // now, we must compute (S^e1)*c2 
      t := dom::qShift(c2, e1);
      t := c1*t;
      if type(r[e1+e2]) = "_index" then 
        r[e1+e2] := normal(t);
      else
        r[e1+e2] := normal(r[e1+e2] + t);
      end_if;
    end_for;
  end_for;
  r := poly([[rhs(t),lhs(t)] $ t in r], [Shift]);
  return(dom::new(r));
end_proc; // _mult2

//------------------------------------------------------------------
// Commute -- compute the commuted expression 
// Shift^e * c = Commute(e, c) * Shift^e
//------------------------------------------------------------------
Commute := proc(e : DOM_INT, c)
//option remember;
begin
  error("unused in q-case");
end_proc; // Commute

//------------------------------------------------------------------
// _divide -- rightdivide p1 by p2
//             i.e., determine q and r such that
//             p1 = p2*q + r
//             where r is of lower or equal degree than p2
//------------------------------------------------------------------
_divide := dom::rightdivide;
rightdivide := proc(p1 : dom, p2 : dom)
local r,quot,p,t,lcp, lcr, lcq,dd;
begin
  //return (FAIL);
  r := p1; // remainder
  quot := dom::zero;  // quotient
  p := p2;
  while r <> dom::zero do 
    dd := degree(r, Shift)-degree(p, Shift);
    if dd < 0 then break end_if;
    [lcp, lcr] := [lcoeff(p, Shift), lcoeff(r, Shift)]; 
    lcp := dom::qShift(lcp, dd); // shift back
    lcq := normal(lcr/lcp);
    t := dom::new(poly([[lcq, dd]], [Shift]));
    quot := quot + t;
    r := mapcoeffs(r - t*p, normal);
  end_while;
  return ([quot,r]);
end_proc; // rightdivide

//------------------------------------------------------------------
// leftdivide -- leftdivide p1 by p2
//                i.e., determine q and r such that
//                p1 = q*p2 + r
//                where r is of lower or equal degree than p2
//------------------------------------------------------------------
leftdivide := proc(p1 : dom, p2 : dom)
local r,quot,p,t,lcp, lcr, lcq,dd,d;
begin
  //return (FAIL);
  r := p1; // remainder
  quot := dom::zero;  // quotient
  p := p2;
  while r <> dom::zero do 
    dd := degree(r, Shift)-degree(p, Shift);
    if dd < 0 then break end_if;
    [lcp, lcr] := [lcoeff(p, Shift), lcoeff(r, Shift)]; 
    d := degree(p, Shift);
    lcq := normal(lcr/lcp);
    lcq := dom::qShift(lcq, -d); // shift back
    t := dom::new(poly([[lcq, dd]], [Shift]));
    quot := quot + t;
    r := mapcoeffs(r - p*t, normal);
  end_while;
  return ([quot,r]);
end_proc; // rightdivide

//------------------------------------------------------------------
// gcrd -- greatest common right divisor
//------------------------------------------------------------------
gcd := dom::gcrd;
gcrd := proc(p1:dom, p2:dom) 
local t, f1, f2, quot, p, tt;
begin
  // monify input
  t := (lcoeff(p1))^(-1);
  p1 := mapcoeffs(p1, c -> normal(c*t));
  t := (lcoeff(p1))^(-1);
  p2 := mapcoeffs(p2, c -> normal(c*t));
  
  [f1, f2] := _if(degree(p1)>degree(p2),[p1,p2],[p2,p1]);
  while TRUE do // exit via break
    [quot, t] := f1/f2;
    t := mapcoeffs(t, normal);
    quot := mapcoeffs(quot, normal);
    if dom::iszero(t) then break end_if; 
    tt := (lcoeff(t))^(-1);
    [f1, f2] := [f2, mapcoeffs(t, c -> normal(c*tt))];
  end_while;
  return (f2);
end_proc; // gcrd

//------------------------------------------------------------------
// lclm -- least common left multiple
//------------------------------------------------------------------
lcm := dom::lclm;
lclm := proc(p1:dom, p2:dom) 
local n,i,E, AA,conv,Ansatz,Sol, d, q1, q2, r1, r2, ind;
begin
  n := degree(p1) + degree(p2);
  ind := ((indets(expr(p1)) union indets(expr(p2))) union {X, q}) minus { Shift };
  repeat
    // now, create an SO which has only variables as coefficients
    AA := [`#`.i $ i=0..n-1];
    Ansatz := dom::new(Shift^n+_plus((AA[i+1] * Shift^i) $ i=0..n-1));
    // symbolically right-divide
    [q1, r1] := (Ansatz/p1);
    [q2, r2] := (Ansatz/p2);
    // now, compare coefficients
    conv := (y) ->  expr(y); //mapcoeffs(expr(y), AA, R);
    E := {(conv(coeff(r1, i))) $ i=0..degree(r1),
          (conv(coeff(r2, i))) $ i=0..degree(r2) };
    // solve the linear equations
    Sol := linsolve(E, AA);
    d := nops(_union(op(map(Sol, indets@rhs))))-nops(ind);
    n := n-d;
  until d = 0 end_repeat;
  dom::new(subs(expr(Ansatz), Sol));
end_proc; // lclm

//------------------------------------------------------------------
// _invert -- inversion generally impossible
//------------------------------------------------------------------
_invert := (p) -> FAIL;

//------------------------------------------------------------------
// adjoint -- ...
//------------------------------------------------------------------
adjoint := proc(p : dom) 
local i;
begin
  _plus(subs(coeff(p,i), X = q^(-i)*X^(-1))*Shift^i $ i=0..dom::order(p));
  dom::new(%); 
end_proc;

//==================================================================
// Accessing coefficients and properties
//------------------------------------------------------------------
// lterm -- return the leading term
//------------------------------------------------------------------
lterm := proc(p)
local i;
begin
  dom::new(lterm(rep(p), args(i) $ i=2..args(0)));
end_proc; // lterm

//------------------------------------------------------------------
// lmonomial -- return the leading monomial
//------------------------------------------------------------------
lmonomial := proc(p) 
local i;
begin
  dom::new(lmonomial(rep(p), args(i) $ i=2..args(0)));
end_proc; // lmonomial

//------------------------------------------------------------------
// lcoeff -- return the leading coefficient
//------------------------------------------------------------------
lcoeff := proc(p)
local i;
begin
  lcoeff(rep(p), args(i) $ i=2..args(0));
end_proc; // lcoeff

//------------------------------------------------------------------
// tcoeff -- return the trailing coefficient
//------------------------------------------------------------------
tcoeff := proc(p)
local i;
begin
  tcoeff(rep(p), args(i) $ i=2..args(0));
end_proc; // tcoeff

//------------------------------------------------------------------
// poly - return the corresponding polynomial
//------------------------------------------------------------------
poly := (p) -> rep(p);

//------------------------------------------------------------------
// degree -- return the degree in Shift
//------------------------------------------------------------------
degree := proc(p) 
local i;
begin
  return (degree(rep(p), args(i) $ i=2..args(0)));
end_proc;

order := degree;

//------------------------------------------------------------------
// nterms -- return the number of terms in Shift
//------------------------------------------------------------------
nterms := (p) -> nterms(rep(p));

//------------------------------------------------------------------
// indets -- return the indeterminats
//------------------------------------------------------------------
indets := (p) -> {};

//------------------------------------------------------------------
// iszero -- check for being zero
//------------------------------------------------------------------
iszero := (p) -> iszero(rep(p));

//------------------------------------------------------------------
// coeff -- find a coefficient
//------------------------------------------------------------------
coeff := proc(p : dom, n) 
begin
  if (args(0) = 2) then
    return (coeff(rep(p), Shift, n));
  elif args(0) = 1 then
    return (coeff(rep(p)));
  elif args(0) = 3 and args(2)=Shift then
    return (coeff(rep(p), Shift, args(3)));
  else
    error("Wrong number of arguments");
  end_if;
end_proc; // coeff

//==================================================================
// Output and Conversion
//------------------------------------------------------------------
// print
//------------------------------------------------------------------
print := proc(p : dom) 
local i, l, St;
begin
//  return(op(rep(p),1)); 
  
  St := (x) ->
      _if(x[2]=0,
        _if(denom(x[1])=1,
          collect(numer(x[1]), X),
          hold(_divide)(collect(numer(x[1]),X),collect(denom(x[1]),X))
        ),
        _if(x[1]=1,
          Shift^x[2],
          hold(_mult)(
            _if(denom(x[1])=1,
              collect(numer(x[1]), X),
              hold(_divide)(collect(numer(x[1]),X),collect(denom(x[1]),X))
            ),
            Shift^x[2]
          )
        )
      );

  l := poly2list(rep(p));
  if nterms(rep(p)) = 0 then
    0
  elif nterms(rep(p))<=1 then
    St(l[1]);
  else
    hold(_plus)(St(i)$ i in l);
  end_if;
end_proc;

//------------------------------------------------------------------
// TeX
//------------------------------------------------------------------
TeX := proc(p : dom)
local i;
begin
  St := (x) ->
      _if(x[2]=0,
        _if(degree(denom(x[1]),X)=0,
          collect(x[1], X),
          hold(_divide)(collect(numer(x[1]),X),collect(denom(x[1]),X))
        ),
        _if(x[1]=1,
          Shift^x[2],
          hold(_mult)(
            _if(degree(denom(x[1]),X)=0,
              collect(x[1], X),
              hold(_divide)(collect(numer(x[1]),X),collect(denom(x[1]),X))
            ),
            Shift^x[2]
          )
        )
      );

  l := poly2list(rep(p));
  if nterms(rep(p)) = 0 then
    "0"
  elif nterms(rep(p))<=1 then
    generate::TeX(St(l[1]));
  else
    l := map(l, (x) -> generate::TeX(St(x)));
    l[1]._concat((" + ",  l[i]) $ i=2..nops(l));
  end_if;
end_proc;

//------------------------------------------------------------------

// expr
//------------------------------------------------------------------
expr :=  (p) -> expr(rep(p));

//------------------------------------------------------------------
// mapcoeffs
//------------------------------------------------------------------
mapcoeffs := proc(p:dom) 
local i;
begin
  mapcoeffs(rep(p), args(i) $ i=2..args(0));
  dom::new(%);
end_proc; // mapcoeffs

//------------------------------------------------------------------
// _mod
//------------------------------------------------------------------ 
modp := dom::_mod;
_mod := proc(r : dom, e : DOM_INT) 
  local p;
  begin
    p := mapcoeffs(rep(r), _mod, e);
    return(dom::new(p)); 
  end_proc; // _power

//------------------------------------------------------------------
// polyfy
//------------------------------------------------------------------
polyfy := proc(p : dom) 
local t1, t2, i, Flg, c;
begin
  Flg := FALSE;
  for i from 0 to dom::order(p) do
    c := expr(dom::coeff(p,i));
    if TRUE <> is(c, PolyExpr(q,X)) then
      Flg := TRUE;
      break;
    end_if;
  end_for;
  if Flg=FALSE then return([p,1]) end_if;
  t1 := lcm(map(coeff(p), (c) -> denom(expr(c))));
  p := mapcoeffs(p, _mult, t1);
  t2 := gcd(map(coeff(p), (c) -> numer(expr(c))));
  p := mapcoeffs(p, (c) -> normal(c/expr(t2)));
  //t2 := 1;
  [p, normal(t1/t2)];
end_proc; // polyfy

//------------------------------------------------------------------
// denombound -- returns a polynomial which is a multiplicative 
//                bound for the denominator of any rational solution
//------------------------------------------------------------------
denombound := proc(p : dom)
local co, i, t, B0, B1, a, b, base, bound, f, lc, tc, r, l1, l2;
save T, L, bc;
begin
  delete T,L;
  lc := dom::qShift(dom::lcoeff(p), -dom::degree(p));
  tc := dom::tcoeff(p);
  // base shall become a list of representatives of the occurring shift-
  // equivalence-classes.
  r := dom::baserep(lc, tc);
  base := r[1];
  [B0, B1] := r[3];
  co := table(L=r[2][1][2], T=r[2][2][2]);
  // co[L] and co[T] contain a list for the leading/trailing coeff resp.
  // each element is a list that stands for one shift-equivalence-class
  // determined by the 1st element, the basevector.
  // within these classes every element is a list of the form
  // [shift, multiplicity] corresponding to a factor 
  // basevector(x-shift)^multiplicity.
  // now, go thru the factors again...
  bound := 1; //r[2][1][1]*r[2][2][1]; 
  for b in base do
    l1 := array(B0[b]..B1[b], [0 $ B1[b]-B0[b]+1]);
    // going up
    f := co[L][b];
    if f = FAIL then next end_if;
    for i from B0[b] to B1[b] do
      t := 0;
      for a in f do
        if i >= a[1] then
          t := t + a[2]; 
        end_if;
      end_for;
      l1[i] := t;
    end_for;
    // going down
    l2 := array(B0[b]..B1[b], [0 $ B1[b]-B0[b]+1]);
    f := co[T][b];
    if f = FAIL then next end_if;
    for i from B1[b] downto B0[b] do
      t := 0;
      for a in f do
        if i <= a[1] then
          t := t + a[2]; 
        end_if;
      end_for;
      l2[i] := t;
    end_for;
    bound := bound * _mult( (dom::qShift(expr(b), i))^min(l1[i], l2[i]) $ i=B0[b]..B1[b] );
  end_for; // shifteq-classes
  return (normal(bound));  
end_proc; // denombound

//------------------------------------------------------------------
// valbound -- returns a table indexed by the factors of the 
//             leading/trailing coefficients. For each base-vector
//             it returns the max. valuation to the left/right
//------------------------------------------------------------------
valbound := proc(p : dom)
local co, B0, B1, b, base, bound, lc, tc, r, zeroiffail;
save T, L;
begin
  delete T,L;
  p := dom::polyfy(p)[1]; // don't care about factor for the moment
  lc := dom::qShift(dom::lcoeff(p), -dom::degree(p));
  tc := dom::tcoeff(p);
  // base shall become a list of representatives of the occurring shift-
  // equivalence-classes.
  r := dom::baserep(lc, tc);
  base := r[1];
  [B0, B1] := r[3];
  co := table(L=r[2][1][2], T=r[2][2][2]);
  // co[L] and co[T] contain a list for the leading/trailing coeff resp.
  // each element is a list that stands for one shift-equivalence-class
  // determined by the 1st element, the basevector.
  // within these classes every element is a list of the form
  // [shift, multiplicity] corresponding to a factor 
  // basevector(x-shift)^multiplicity.
  // now, go thru the factors again...
  bound := table();
  zeroiffail := (x) -> _if(x=FAIL, 0, _plus(op(map(x, op, 2))));
  for b in base do
    bound[b] := (-zeroiffail(co[L][b]))..zeroiffail(co[T][b]); 
  end_for; // shifteq-classes
  return (bound);  
end_proc; // valbound

//------------------------------------------------------------------
// baserep -- computes the representation of one or more polynomials
//             wrt. a base of non-shift-eq. polys.
// returns [base, L, [B0,B1]]
// where base is the used base, B0 and B1 are the minimal/maximal
// used shifts respectively and L is a List containing a list for
// each argument. 
// example: 
// baserep(SR::baserep(poly(3*(x+1)*(x-2)*(x^2+1)*((x + 3)^2 + 1)), poly(x-29)))
// yields
// [ [x - 2, x^2 + 1], 
//   [ [3, table((x^2 + 1)=[[0, 1], [3, 1]],
//               (x - 2)=[[0, 1], [3, 1]])], 
//     [1, table((x^2 + 1)=FAIL,
//               (x - 2)=[[-27, 1]])]
//   ], 
//   [-27, 3] 
// ]
// (poly() removed for simplicity)
// Hoeij uses an actual zero of the irreducible factors, so his
// 0 corresponds to x here.
//------------------------------------------------------------------
findshift := proc(p1, p2) 
local i, quot, l1, l2, k, c1, c2, t1, t2;
// option remember;
begin 
  p1 := poly(p1,[X]);
  p2 := poly(p2,[X]);
  if degree(p1) <> degree(p2) then return(FAIL) end;
  if nterms(p1) <> nterms(p2) then return(FAIL) end;
  if iszero(p1) and iszero(p2) then return(0) end;
  t1 := tcoeff(p1);
  p1 := multcoeffs(p1, 1/t1);
  t2 := tcoeff(p2);
  p2 := multcoeffs(p2, 1/t2);
  l1 := lcoeff(p1);
  l2 := lcoeff(p2);
  quot := normal(l2/l1);
  if quot=1 then 
    return(_if(bool(p1=p2), 0, FAIL)); 
  elif quot=q then 
    k := 1;
  elif type(quot)="_power" and op(quot,1)=q and testtype(op(quot,2)/degree(p1), Type::Integer) then
    // may be shift-equivalent
    k := op(quot,2)/degree(p1);
  else
    return(FAIL);
  end_if;
  for i from 0 to degree(p1) do
    c1 := coeff(p1,i);
    c2 := coeff(p2,i);
    if iszero(c1) and iszero(c2) then next end;
    if iszero(c1) or iszero(c2) then return(FAIL) end;
    if not iszero(normal(c2 - c1*q^(i*k))) then return(FAIL) end;
  end_for;
  return(k)
end_proc;

baserep := proc()
local Rep, B0, B1, b, base, cc, cn, co, i, j, t, t2, t3, w, s, findshift, tc, ct;
// option remember;
begin
  P := map([args()], (x) -> poly(x,[X]));
  tc := map(P, tcoeff);
  P := [ multcoeffs(P[i], 1/tc[i]) $ i=1..args(0) ];
  P := map(P, factor);
  ct := map(P, op, 1); // [op(P,1)[1], op(P,2)[1]];
  ct := zip(tc, ct, _mult);
  findshift := dom::findshift;
  //findshift := proc() option remember; begin dom::findshift(args()) end_proc;
  // base shall become a list of representatives of the occurring shift-
  // equivalence-classes.
  base := [op(Factored::factors(P[i])) $ i=1..args(0)];  
  //base := _concat(op(map(P, (y) -> _if(nops(y) = 1, [1], Factored::factors(y)))));
  for i from 1 to nops(base) do
    if base[i] = FAIL then next end_if;
    for j from i+1 to nops(base) do
      if base[j] = FAIL then next end_if;
      if ((s := findshift(base[i], base[j]))) <> FAIL then
        // make sure we have only negative shifts
        if s < 0 then
          base[i] := base[j];
        end_if;
        base[j] := FAIL;
      end_if;
    end_for; // i base
  end_for;
  base := select(base, _unequal, FAIL);
  // Rep will contain a list for each argument
  // each element is a list that stands for one shift-equivalence-class
  // determined by the 1st element, the basevector.
  // within these classes every element is a list of the form
  // [shift, multiplicity] corresponding to a factor 
  // basevector(x-shift)^multiplicity.
  B0 := table(b = infinity $ b in base);
  B1 := table(b = -infinity $ b in base);
  Rep := [];
  for w in P do 
    co := [ op(w) ];
    cn := (nops(co)-1)/2;
    cc := co[1];
    t := [ [co[2*i], co[2*i+1]] $ i=1..cn ];
    co := table();
    for b in base do
      t2 := [];
      for j from 1 to nops(t) do
        if t[j] = FAIL then next end_if;
        if FAIL <> (t3 := dom::findshift(b, t[j][1])) then
          t2 := t2.[ [t3, t[j][2]] ];
          B0[b] := min(B0[b], t3);
          B1[b] := max(B1[b], t3);
          t[j] := FAIL;
        end_if; // same shifteq-class
      end_for; // list of factors
      //if nops(t2) > 0 then
        co[b] := t2;
      //else
      //  co[b] := FAIL;
      //end_if;
    end_for; // list of shifteq-classes
    Rep := Rep.[ [cc, co] ];
  end_for; // args
  return([base, Rep, [B0, B1]]);
end_proc; // baserep

//------------------------------------------------------------------
// symprod(p,q) -- compute the symmetric product of p and q 
//                  where q
//                  has order 1.
//                  cf. CluHoe05, p5
//------------------------------------------------------------------
symprod := proc(p1 : dom, p2 : dom) 
local r, P, i, n, B, j;
save a,b,c;
begin
  if dom::degree(p2) = 1 then 
    n := degree(p1);
    r := -coeff(p2,0)/coeff(p2,1);
    B := array(0..n, [(coeff(p1,i) * _mult((expr(r) | X=q^j*X) $ j=i..n-1) ) $ i=0..n-1, coeff(p1,n)]);
    P := dom::new(_plus( B[i]*Shift^i $ i=0..n ) );
    return(P);
  else
    error("not yet implemented");
    /*delete a,b,c;
    f := op(p,1);
    g := op(q,1);
    m := degree(f);
    n := degree(g);
    s1 := -(f(a)-lmonomial(f(a),a))/lcoeff(f);
    s2 := -(g(b)-lmonomial(g(b),b))/lcoeff(g);
    F := poly(_plus(c.i*(a*b)^i $ i=0..m*n), [a,b]  );
    while degree(F,a) >= m or degree(F,b) >= n do
      F2 := poly(0, [a,b]);
      for t in poly2list(F) do
        [da, db] := t[2];
        poly(t[1] * _if(da >= m, a^(da-m)*s1, a^da) * _if(db >= n, b^(db-n)*s2, b^db), [a,b]);
        F2 := F2 + %;
      end_for;
      F := F2;
    end_while;  
    L := { t[1] = 0 $ t in poly2list(F) };
    G := subs(_plus(c.i*X^i $ i=0..m*n), op(linsolve(L)));
    G := subs(G, { i=1 $ i in (indets(G) minus {X}) });
    poly(G,[x]);*/
  end_if;
end_proc; // symprod

//------------------------------------------------------------------
// polysols -- returns the polynomial solutions of the SO
//------------------------------------------------------------------
qupperbound := proc(L : dom)
local a,m,n,p,k,i,z,s,t,alpha;
begin  
  n := dom::order(L);
  a := [expand(expr(dom::coeff(L,i))) $ i=0..n];
  m := max(op(map(a, (F) -> _if(has(F, X), degree(F, X), 0)))); // maximum of the degrees of the coefficent polynomials
  alpha :=  map(a, (F) -> _if(has(F, X), coeff(F,[X],m), F)); 
  p := poly(_plus(alpha[k+1] * X^k $ k=0..n),[X],Expr);
  s := ldegree(p);
  z := poly(alpha[s+1], [q]);
  if z=FAIL then return(-1) end;
  for i from ldegree(z) downto 0 do 
     if expand(p(q^i))=0 then 
        JJJJ := JJJJ + 1;
        return(i);
     end_if; 
  end_for;
  -1
end_proc;

qpolysols := proc(L:dom, isPoly=FALSE, bound=FAIL)
local i,n,N,pol,eq,sol,Y;
save a;
begin
  delete a;
  if not isPoly then
    L := dom::polyfy(L)[1];
  end_if;
  Y := indets(expr(L));
  n := dom::order(L);
  if bound = FAIL then
    N := dom::qupperbound(L);
  else 
    N := bound;
  end_if;
  //print(N);
  if N<0 then 
     return({});
  end_if;
  pol := _plus(`#`.i*X^i $ i=0..N);
  eq := [ coeff(collect(L(pol), X), [X]) ];
  //warning("halt!");
  if FALSE and has(eq, q) then
    sol := liinalg::linsolve(eq, [`#`.j $ j=0..N]);
    sol := map(sol, (A)->normal(subs(pol, A)));
  else 
    sol := numeric::linsolve(eq, [`#`.j $ j=0..N], Symbolic);
    sol := subs(pol, sol);
    XX := indets(sol) minus ({X, Shift, q} union indets(expr(L)));
    sol := map(XX, x -> normal(subs(sol, [x=1, (Y=0 $ Y in XX minus {x})])));
  end_if;
  select(sol, (A) -> not iszero(A));
end_proc; // qpolysols

//------------------------------------------------------------------
// ratsols -- returns the rational solutions of the SO
//------------------------------------------------------------------
ratsols := proc(p : dom)
local D, B, n, P, i, Y, pSol, rSol;
begin
  //Y := indets(expr(p)) minus {X, Shift};
  //if nops(Y) > 0 then
  //  t := dom::coeff(p);
  //  for i in Y do
  //    i^max(map(t, (x) -> degree(denom(normal(expr(x))),[op(Y)],i)));
  //    p := dom(%)*p;
  //  end_for;
  //end_if;
  p := dom::polyfy(p)[1];
  D := dom::denombound(p); // Denominator
  n := degree(p);
  B := array(0..n, [ coeff(p, i) / (D | X=q^i*X) $ i=0..n ]);
  P := dom::new(_plus( expr(B[i])*Shift^i $ i=0..n ));
  P := dom::polyfy(P)[1];
  Y := genident("Y");
  pSol := dom::qpolysols(P);
  rSol := [FAIL $ i=1..nops(pSol)];
  for i from 1 to nops(pSol) do
    P := pSol[i];
    Y := indets(P);
    rSol[i] := P/D;
  end_for;
  map({op(rSol)}, normal);
end_proc; // ratsols

myIsInteger := proc(e, epsilon=0.01)
local i;
begin
  for i in float(e) do
    if domtype(i)<>DOM_FLOAT and domtype(i)<>DOM_COMPLEX then
      return(TRUE);
      print(e);
      print(float(e));
      print(i);
    end_if;
    if abs(Im(i)) < 0.1 then
      if abs(i-round(i)) < 99 then // epsilon then
        return(TRUE)
      end_if;
    end_if;
  end_for;
  return(FALSE);
end_proc;


//------------------------------------------------------------------
// factorHoeij -- returns right-factors of order 1
//                This is van Hoeij's algorithm
//------------------------------------------------------------------
factorHoeij := proc(L : dom, opt=FAIL)
local i, L2, RF, SS, c, cand, d, l, pai, r, s, sings, sols, t, c0, pa0,
  vb, S, va, J, JJ, FE, RS, r1, r2, R2, tmp, rn, rd, cc, S0, C0, CC0,
  lcn, lcd, tnc, tcd, Pa0, x;
begin
  ST := 0;
  // we assume that there are no rational or polynomial solutions
  L := dom::polyfy(L)[1];
  vb := [op(dom::valbound(L))];
  if FALSE and 100 < _mult(op(rhs(i),2)-op(rhs(i),1) $ i in vb) then
    // evil heuristic: The problem-points don't change if we compute the lclm
    // with an Shift-someinteger, but there is a hope that the candidate-types 
    // become fewer.
    // there are so many possibilities that the lclm-trick is worth the efford
    L2 := dom::lclm(L, dom::new(Shift-3));
    va := dom::valbound(L2);
    t := combinat::cartesianProduct::generator({op(vb)}, {op(va)});
    vb := table();
    while FAIL <> (i := t()) do
      if FAIL <> polylib::shiftequiv(lhs(i[1]),lhs(i[2])) then
        r1 := rhs(i[1]); r2 := rhs(i[2]);
        r := max(op(r1,1), op(r1,1))..min(op(r1,2), op(r1,2));
        if op(r,1) <= op(r,2) then
          vb[lhs(i[1])] := r;
        end_if;
      end_if;
    end_while;
  end_if;
  S0 := {0};
  pa0 := dom::localTypeAt0(L):
  Pa0 := table(op(x,2) = {} $ x in pa0);
  for x in pa0 do
    if not iszero(x) then
      Pa0[op(x,2)] := Pa0[op(x,2)] union { normal(op(x,1)) }
    end_if
  end_for;
  S0 := { op(map(pa0, op, 2)) };
  //print(C0);
  //CC0 := map(C0, (a)->_if(0=traperror(subs(a, q=1)),eval(subs(a,q=1)),null())):
  pai := dom::localTypeAtInifinity(L):
  //print(vb);
  sings := [lhs(t) $ t in vb, poly(X, [X])];
  sols := {};
  J := 0; JJ := 0;
  J := nops(pai) * _mult(1+op(rhs(t),2)-op(rhs(t),1) $ t in vb);
  userinfo(2, expr2text(J)." Versuche insgesamt.");
  for l in pai do
    [c,s,d] := l;
    //print(l);
    if TRUE <> testtype(expand(c), Type::RatExpr(q)) then 
      userinfo(9, "c=".expr2text(c)." not rational, damn!"); //substituting q=1.");
      //next;
      c := subs(c, q=1); 
    end_if;
    cand := dom::cgenerator({i$i=rhs(t)} $ t in vb, S0);
    while FAIL <> (t := cand()) do
      J := J-1;
      if s+_plus(degree(sings[i])*t[i] $i=1..nops(t)) <> 0 then 
        userinfo(3, "Kombination ".J." wegen 1ster Fuchs-Relation in unendlich übersprungen.");
        next 
      end_if; // 1st fuchs-relation
      if not _plus(ldegree(sings[i])*t[i] $i=1..nops(t)) in S0 then 
        userinfo(3, "Kombination ".J." wegen 1ster Fuchs-Relation in 0 übersprungen.");
        next 
      end_if; // 1st fuchs-relation
      C0 := Pa0[_plus(ldegree(sings[i])*t[i] $i=1..nops(t))];
      JJ := JJ+1;
      userinfo(3, "Schon ".JJ." Kombinationen getestet.");
      // now, we know that both fuchs-relations hold and we construct an Operator of 
      // appropriate type
      if (domtype(c) in {DOM_INT, DOM_RAT, F}) or F=Dom::ExpressionField() then 
        FE := F;
        RS := dom;
      else
        FE := domtype(c); // we possibly need to deal with field-extensions
        RS := Dom::ShiftOperators(X, Shift, FE);
      end_if;
      
      lcn := _mult(lcoeff(sings[i])^max(0,t[i]) $ i=1..nops(t));
      lcd := _mult(lcoeff(sings[i])^(-min(0,t[i])) $ i=1..nops(t));
      tcn := _mult(tcoeff(sings[i])^max(0,t[i]) $ i=1..nops(t));
      tcd := _mult(tcoeff(sings[i])^(-min(0,t[i])) $ i=1..nops(t));
            
      c0 := normal(c/lcn*lcd*tcn/tcd);
      //print()
      //print([s, c, c0, r]);
      if not c0 in C0 then  // traperror(subs(c0,q=1))=1 and not eval(subs(c0,q=1)) in CC0 then 
        userinfo(3, "c0=".expr2text(c0)." nicht in ".expr2text(C0)); 
        for i in C0 do
          tmp := normal(expand(i/c0));
          //print(c0, i, tmp, type(tmp));
          if has(tmp, q) and testtype(tmp, Type::RatExpr(q))then
            if tmp = q or type(tmp) = "_power" or type(-tmp) = "_power" then 
              c0 := FAIL; 
              break; 
            end;
          end_if;
        end_for;
        //print(Unquoted, "c0=".expr2text(c0)); 
        if not c0 = FAIL then 
          userinfo (1, "Skipping.");
          //print(".");
          next; 
        end_if; 
      end_if;
      
      r := normal(expr(_mult(sings[i]^-t[i] $ i=1..nops(t)))*lcn/lcd/c);
      //print(r);
      if type(c) = RootOf then warning("damn!"); c := 1; end_if;
      userinfo(2, "c=".expr2text(c)." s=".expr2text(s)." r=".expr2text(r)." cc=".expr2text(cc));
      RF := RS(Shift-r);
      L2 := RS::symprod(RS(L), RF);
      st := time();
      SS := RS::ratsols(L2);
      //print(L2);
      ST := ST + time()-st;
      if nops(SS) > 0 then
        for S in SS do 
          if nops((va := indets(SS) minus {X, q})) > 0 then
            S := subs(S, i=1 $ i in va);
          end_if;
          if iszero(S) then
            next;
          else
            tmp := RS(Shift - subs(S, X=q*X)/S/expr(r));
            userinfo(3, "Lösung gefunden: ".expr2text(tmp));
            if opt=hold(PrincipalValue) then
              return(tmp);
            end_if;
            sols := sols union { tmp }; //{ [tmp, s, c, c0, rn, rd] };
          end_if;
        end_for
      end_if;
    end_while;
  end_for;
  userinfo(3, "Gesamtkombinationen probiert: ".JJ);
  return(sols);
end_proc;


//------------------------------------------------------------------
// factorOrder2 -- 
//------------------------------------------------------------------
rightFactorOrder := proc(L : dom, ord)
local L2, P1, SS, Shift_on_b, W, b, b2, bT, bb, bn, bt, bv, c, cv, 
  eP, eq, fl, i, j, l, leftfactors, m, mop, mopo, rp, s, sgn, sol, 
  vrs, vv, yv, invShift_on_b, vlen, mi, ma, Sol, bbx, einsinv;
save S, rho, t;
begin
  delete S, rho, t;
  userinfo(1, "called rightFactorOrder");

  m := dom::order(L);

  // Handle some special cases
  if m = ord then
    return({L});
  elif m < ord then
    warning ("Operator of order ".m." won't have factors of order ".ord.".");
    return({});
  elif ord = 1 then
    return({dom::qPetkovsek(L, PrincipalValue)});
  elif ord > m/2 then
    return(map(dom::rightFactorOrder(dom::adjoint(L), m-ord), dom::adjoint));
  end_if;

  // initialize
  mopo := binomial(m,ord);
  L := mapcoeffs(L, _divide, lcoeff(L));

  c := array(0..m, [expr(coeff(L,i)) $ i = 0..m]);
  s := array(0..m-1, [`#shift^`.i $i=0..m-1]);
  eP := map(combinat::choose({i $ i=0..m-1}, m-ord), (x)->[op(x)]);
  mopo := nops(eP);
  bT := table((eP[i+1] = `#b`.i) $ i=0..mopo-1);
  //bT := table((i = `#`._concat("s".expr2text(i[j])."^" $ j=1..m-3).i[m-2]) $ i in eP );
  bb := array(0..mopo-1, [rhs(b) $ b in bT]);
  
  //
  // Preparations to handle arbitrary order
  //
  // Compute the action of Shift on the base-vectors of the external product
  //
  Shift_on_b := [ X = q*X ];
  for b in bT do
    bn := rhs(b);
    bv := lhs(b);
    b2 := map(bv, _plus, 1);
    if type(bT[b2]) <> "_index" then
      // lucky case, shift maps b to the element b2
      Shift_on_b := Shift_on_b . [ bn = bT[b2] ];
      next;
    end_if;
    // complicated case, shift maps b to a linear combination
    bt := 0;
    for l in [b2[1..-2].[j] $ j in select([i $ i=0..m-1], (x)->not(has(b2[1..-2], x)))] do
      sgn := (-1)^(1+nops(l)-contains(sort(l),l[-1]));
      bt := bt + sgn * c[l[-1]]*bT[sort(l)];
    end_for;
    Shift_on_b := Shift_on_b . [ bn = bt ];
  end_for;
  //
  // What does Shift^-1 do on an Vector?
  //
  invShift_on_b := [ X = X/q ];
  einsinv := poly(_plus(subs(-1/c[0], X=X/q)*S^(m-1) , -subs(c[i+1]/c[0], X=X/q)*S^i $ i=0..m-2), [S], Expr);
  for b in bT do
    bn := rhs(b);
    bv := lhs(b);
    b2 := map(bv, _plus, -1);
    if type(bT[b2]) <> "_index" then
      // lucky case, shift maps b to the element b2
      invShift_on_b := invShift_on_b . [ bn = bT[b2] ];
      next;
    end_if;
    // complicated case, shift maps b to a linear combination
    bt := 0;
    for l in [[j].b2[2..-1] $ j in select([i $ i=0..m-1], (x)->not(has(b2[2..-1], x)))] do
      t := sort(l);
      sgn := (-1)^(1+contains(t,l[1]));
      bt := bt + sgn * coeff(einsinv, l[1])*bT[t];
    end_for;
    invShift_on_b := invShift_on_b . [ bn = bt ];
  end_for;
  //return([Shift_on_b, invShift_on_b]);
  //
  // find how to get the coefficients of an 2nd-order-factor from the eigenvector
  //
  // assume the ord-th-order factor is 
  // Shift^ord + rho[ord-1]*Shift^ord[-1] + ... + rho[0]
  // then, the eigenvector has the form
  // (Shift^ord + rho[ord-1]*Shift^(ord-1) + ... + rho[0]) 
  //   ^ (Shift^(ord+1) + rho[ord-1]*Shift^(ord) + ... + Shift*rho[0])
  //     ^ ... ^ (Shift^(m-1) + rho[ord-1]*Shift^(m-2) + ... + Shift^(m-ord)*rho[0]) 
  SS := [ S[i] $ i=0..m-1 ];
  W := poly(_mult( _plus(_if(j=ord,1,rho[j])*S[j+i] $ j=0..ord) $ i=0..m-(ord+1) ), SS);
  W := select(poly2list(W), (x) -> _lazy_and(max(op(op(x,2)))<2, type(op(x,1)) in {"_index", DOM_INT} ));
  if nops(W) <> ord+1 then error("This ought to work!") end_if;
  W := table( (t[1]=bT[select([i $ i=0..m-1], (i)->(t[2][i+1]=1))]) $ t in W );
  // end preparations for arbitrary order
  
  userinfo(1, "initialization done.");
  
  // Compute a cyclic vector and the corresponding minimal operator
//  MOP := [0,0,0,0];
  S := {};
  bbx := [op(bb), X];
  i := 0;
    Sol := {};
//  for J from 1 to 4 do
    repeat
      userinfo(1, "trying to find a cyclic vector ".expr2text(i));
      // first, choose a vector that shall be cyclic. 
      repeat
        i := i+1;
        if i<mopo then
          cv := bb[i-1];
        elif i < mopo+mopo^2 then
            cv := bb[i-mopo mod mopo]+X*bb[i-mopo div mopo];
        else
          t := i-(mopo+mopo^2)+1;
          rp := (x) -> expr(polylib::randpoly([X], Degree=t+1, Terms=t, Coeffs=random(-t..t), Monic));
            cv := _plus(bb[j]*rp() $ j=0..mopo);
        end_if;
      until not cv in S end_repeat;
      cv := poly(cv, bbx);
      if cv in S then next end_if;
      S := S union {cv};
      // compute the iterated Shifts on the cyclic vectors
      vv := array(-mopo..mopo, [0 $ mopo, cv, 0 $ mopo]);
      vlen := array(-mopo..mopo, [0 $ mopo, length(cv), 0 $ mopo]);
      for j from 1 to mopo do
        vv[j] := poly(expand(eval(subs(expr(vv[j-1]), Shift_on_b))), [op(bb)]);
        if vv[j] in S then next end_if;
        S := S union {vv[j]};
        vlen[j] := length(vv[j]);
      end_for;
      for j from -1 downto -mopo do
        vv[j] := poly(expand(eval(subs(expr(vv[j+1]), invShift_on_b))), [op(bb)]);
        if vv[j] in S then next end_if;
        S := S union {vv[j]};       
        vlen[j] := length(vv[j]);
        if vlen[j] > vlen[j+mopo+1] then break; end_if;
      end_for;
      mi := 0; ma := 0;
      while ma-mi < mopo do
        if vlen[ma+1] <= vlen[mi-1] then
          ma := ma + 1;
        else
          mi := mi - 1;
        end_if;
      end_while;
      vv := array(0..mopo, [vv[i] $ i=0..mopo]);
      //vv := array(0..mopo, [vv[i] $ i=mi..ma]);
      cv := vv[0];
      // set up the equations and solve the linear systems
      vv := map(vv, op, 1);
      yv := array(0..mopo, [genident("Y") $ i=0..mopo]);
      eq := [ coeff(_plus(yv[j]*vv[j] $ j=0..mopo), [op(bb)]) ];
      sol := solve(eq, [op(yv)], IgnoreSpecialCases);
      mop := subs(_plus(yv[j]*Shift^j $ j=0..mopo), op(sol,1));
      vrs := indets(sol) minus { op(yv), X, Shift, q };
      userinfo(1, "Linear System leads to ".expr2text(nops(vrs))." free parameter(s).");
      // set up the equations and solve the linear systems
    until nops(vrs)=1 and coeff(mop,Shift,0) <> 0 end_repeat;
    mop := dom::new(subs(mop, [vrs[1]=1]));
    // now, cv is a cyclic vector and 
    // mop the corresponding minimal operator of order 6
  userinfo(1, "Found cyclic vector and computed minimal operator.");
  userinfo(1, mop);

  // find 1st-order left(!)factors
  //
  // try factoring the reduced minimal operator
  // 
  leftfactors := dom::qPetkovsek(dom::adjoint(mop));
  userinfo(1, "Found left-factors: ".expr2text(leftfactors));
  if nops(leftfactors)=0 then return ({}); end_if;

  
  for fl in leftfactors do
    // leftdivide by this factor
    P := dom::leftdivide(mop, dom::adjoint(fl));
    // check it's all right
    if not iszero(P[2]) then error("can-not-happen-error"); end_if;
    P := P[1];
  
    // Now, we know that P(cv) is an eigenvector of Shift in the 
    // 2nd external product.
    P1 := poly((_plus(coeff(P,i) * vv[i] $ i=0..mopo-1)), [op(bb)]);
//    beta := map(array(0..mopo-1, [coeff(P1, b.i) $ i=0..mopo-1]), poly, bb);
//    if 0=degree(beta[0]*beta[5]+beta[2]*beta[3]-beta[1]*beta[4]) then
    // cv is an eigenvector that corresponds to a 2nd-order factor.
    t := expr(coeff(P1, W[1], 1));
    L2 := dom::new(Shift^ord + _plus(expr(coeff(P1, W[rho[i]], 1))/t*Shift^i $ i=0..ord-1));
    if iszero(dom::rightdivide(L,L2)[2]) then
      Sol := Sol union { L2 };
    end_if;
  end_for;
  return (Sol);

end_proc; // factorOrder2test



//------------------------------------------------------------------
// factor -- completely factor operators of order 4, try the best 
//           for lower order
//------------------------------------------------------------------
factor := proc(L : dom)
local lhf, order, rhf;
begin
  // there's nothing to do.
  if order(L) = 0 or order(L) = 1 then
    return([L])
  end_if;
  
  // right-factors of order 1
  rhf := dom::factorOrder1(L);
  if nops(rhf) > 0 then
    rhf := op(rhf, 1);
    lhf := dom::rightdivide(L, rhf);
    if not iszero(lhf[2]) then error("Found wrong factor"); end_if; 
    return(dom::factor(lhf[1]) . [rhf]);
  end_if;

  // left-factors of order 1
  lhf := dom::factorOrder1(dom::adjoint(L));
  if nops(lhf) > 0 then
    lhf := dom::adjoint(op(lhf, 1));
    rhf := dom::leftdivide(L, lhf);
    if not iszero(rhf[2]) then error("Found wrong factor"); end_if; 
    return([lhf] . dom::factor(rhf[1]));
  end_if;
  
  // order 3 is irreducible if there are no order 1 factors
  if dom::order(L) = 3 then
    return([L]);
  end_if;

  // we don't know how to deal with higher order
  if dom::order(L) > 4 then
    warning("Result is possibly reducible, we just don't know.");
    return([L]);
  end_if;
  
  // at this point, L has order 4 and no order 1 left- or rightfactors
  
end_proc; // factor

//------------------------------------------------------------------
// polygon -- compute the newton-polygon 
//             returns the integer slopes of the Newton-Polygon and
//             the corresponding Newton-Polynomials
//             cf. CluHoe05, p13
//------------------------------------------------------------------
polygon_alternativ := proc(p : dom, valuation = ((y) -> -degree(y, [X])))
local a, i, m, n, s, v, Res, cc;
begin
  n := dom::degree(p);
  p := dom::polyfy(p)[1];
  if n=0 then return({}) end_if;
  a := array(0..n, [coeff(p, i) $ i=0..n ]);
  v := array(0..n, [_if(iszero(a[i]), infinity, valuation(a[i])) $ i=0..n]);
  s := max( (v[n]-v[i])/(n-i) $ i=0..n-1 );
  m := 0;
  while (v[n]-v[m])/(n-m) <> s do m := m+1 end_while;
  Res := {};
  if TRUE = is(s, Type::Integer) then
    cc := array(0..n-m, [coeff(a[n-i], X, abs(s*i-v[n])) $ i=0..n-m ]);
    P :=  [ [q^(-valuation(x)*((n-m-i)^2-(n-m-i))/2*s) * cc[i], n-m-i] $ i=0..n-m ];
    P :=  poly(P, [X]);
    //P := dom::new(_plus( _if(s*i-v[n]<0,0,coeff(a[n-i], X, s*i-v[n]))*Shift^(n-m-i) $ i=0..n-m ), [X]);
    Res := {[s,P]};
  end_if;
  return(Res union dom::polygon(dom::new(_plus(a[i]*Shift^i $ i=0..m)), valuation));
end_proc; // polygon

polygon := proc(L : dom, valuation = ((y) -> -degree(y, [X])))
local i, Polygon, c, j, n, pol, s, t;
begin
  i := 0;
  n := dom::order(L);
  Polygon := {};
  while i < n do
    if iszero(dom::coeff(L, i)) then i := i+1; next; end_if;
    c := i+1;
    s := infinity;
    for j from i+1 to n do
      if iszero(dom::coeff(L, j)) then next end_if;
      t := (valuation(coeff(L, j))-valuation(coeff(L, i)))/(j-i);
      if t < s then
        s := t;
        pol := eval(subs(coeff(dom::coeff(L, i), X, abs(valuation(coeff(L, i)))), X=1));
      end_if;
      if t = s then
        pol := pol + q^(-valuation(x)*((j-i)^2-(j-i))/2*s)*eval(subs(coeff(dom::coeff(L, j), X, abs(valuation(coeff(L, j)))), X=1))*X^(j-i);
        c := j;
      end_if;
    end_for;
    Polygon := Polygon union { [s, poly(pol, [X])] };
    i := c;
  end_while;
  Polygon;
end_proc;

//------------------------------------------------------------------
// localTypeAtInifinity -- 
//                         cf. CluHoe05, p14
//------------------------------------------------------------------
elt := proc(p : dom)
local pai, pa0, R, i;
begin
  pai := dom::localTypeAtInifinity(p);
  pa0 := dom::localTypeAt0(p);
  R := map(pai, (i) -> [Infinity, i[2], i[1]]), map(pa0, (i) -> [0, i[2], i[1]]);
  return(R, [Infinity, RR::factorHoeij(p)], [0, RR::factorHoeij2(p)]);
end_proc;

localTypeAtInifinity := proc(p : dom) 
local A, L, Polygon, SR, c, f, s, t, v, FE, r, R2, x, C, P;
begin
  A := [];
  Polygon := dom::polygon(p, ((y) -> -degree(y, [X])));
  userinfo(1, "Called Local type at infinity"):
  for t in Polygon do
    [s, P] := t;
    if domtype(s) <> DOM_INT then next end;
    v := -s;
    P := factor(P);
    for f in Factored::factors(P) do
      userinfo(2, "s= ".expr2text(s)." f=".expr2text(f));
      if (degree(f) = 1) then 
        // don't need to take an extension
        FE := F;
        SR := dom;
        R2 := id;
        L := p;
        C := { -f(0)/lcoeff(f) }; // root of the linear factor
      else 
        C := solve(expr(f)=0, X, IgnoreSpecialCases);
      end_if;
      r := 0;
      A := A . [[c,v,r] $ c in C]
    end_for; // s sin sols
  end_for;
  return(A);
end_proc; // localTypeAtInifinity

localTypeAt0 := proc(p : dom) 
local A, L, Polygon, SR, c, f, s, t, v, FE, r, R2, x;
begin
  A := [];
  Polygon := dom::polygon(p, ((y) -> ldegree(y, [X])));
  userinfo(1, "Called Local type at zero"):
  for t in Polygon do
    [s, P] := t;
    if domtype(s) <> DOM_INT then next end;
    v := -s;
    P := factor(P);
    for f in Factored::factors(P) do
      userinfo(2, "s= ".expr2text(s)." f=".expr2text(f));
      //f := poly(subs(expr(f), q=1), [X]);
      //print(Unquoted, " f(q=1)=".expr2text(f));
      if (degree(f) = 1) then 
        // don't need to take an extension
        FE := F;
        SR := dom;
        R2 := id;
        L := p;
        C := { -f(0)/lcoeff(f) }; // root of the linear factor
      else 
        C := solve(expr(f)=0, X, IgnoreSpecialCases);
      end_if;
      r := 0;
      A := A . [[c,v,r] $ c in C];
    end_for; // s sin sols
  end_for;
  return(A);
end_proc; // localTypeAt0


//------------------------------------------------------------------
// toRec -- returns the corresponding expressen for the  rec
//------------------------------------------------------------------
toRec := proc(p : dom, Y)
local i;
begin
  _plus( (expr(coeff(p, i)) * Y(X+i)) $ i = 0..degree(p) )
end_proc; // toRec

//------------------------------------------------------------------
// func_call -- evaluate p(a)
//------------------------------------------------------------------
func_call := proc(p, f, y) 
local m,l;
begin
  f := context(f);
  if args(0) = 3 then
    y := context(y);
  end_if;
  if args(0) < 2 then 
    error("two arguments expected");
  elif args(0) = 2 then
    y := X;
  end_if;
  if domtype(p) <> dom then
    error("Wrong kinds of arguments");
  end_if;
  if (domtype(f) in {DOM_EXPR, DOM_IDENT}) and (has(f,y)) then
    //f := fp::unapply(f, y);
    13
  end_if;
  l := poly2list(extop(p, 1));
  normal(expr(_plus( (m[1] * subs(f, X=q^m[2]*X)) $ m in l)));
end_proc; // func_call

// for speed, faster binomials
mybinom := proc(n,k) local i; option remember; begin 
//  _if(k>(n-k), 
//    _mult(i $ i=k+1..n)/_mult(i$i=2..n-k), 
//    _mult(i $ i=n-k+1..n)/_mult(i$i=2..k)) 
  _if(k<=(n-k), 
    _mult(i $ i=k+1..n)/_mult(i$i=2..n-k), 
    dom::mybinom(n, n-k)); 
end_proc;

//------------------------------------------------------------------
// qShift(e, k) -- shift e by q^k
//------------------------------------------------------------------
qShift := proc(e, k:DOM_INT)
//local x;
begin
  if domtype(e) = DOM_POLY and op(e,2) = [X] then
    poly(map(poly2list(e), x -> [collect(q^(k*x[2])*x[1], q),x[2]]),[X],op(e,3));
  else
    normal(collect(subs(e, X=(q^k*X)), X));
  end_if;
end;

trivfactor := proc(L : dom)
local p, f, l, ll, i, j, LL, pp, eqs, vars, t, e, fl, ii, Sol, e2, per, ret, newvars, tc, Y;
begin
    p := poly(op(L), [Shift]);
    //if lcoeff(p) <> 1 then return(FAIL) end_if;
 //   if 0 <> traperror((
        f := factor(subs(p, q=1));
        l := [];
        fl := [op(f)];
        fl := [ (fl[2*i] $ j=1..fl[2*i+1]) $ i=1..nops(fl)/2 ];
        Y := table();
        for i from 1 to nops(fl) do
            f := fl[i];
            if degree(f, Shift) > 0 then
              e2 := degree(expr(f), X);
              for ii from 0 to degree(f) do
                for j from 0 to e2 do
                  Y[i, ii, j] := `#`.i.`_`.ii.`_`.j.`#`;
                end_for;
              end_for;
              ll := poly2list(f);
              l := l . [_plus(_plus(Y[i, t[2], ii] * X^ii $ ii=0..degree(coeff(expr(f), Shift, t[2]), X))*Shift^t[2] $ t in ll) ];
            else
              error("Illegal Input!");
            end_if;
        end_for;
        tc := [ op(factor(tcoeff(L))) ];
        tc := [ (tc[2*i] $ j=1..tc[2*i+1]) $ i=1..nops(tc)/2 ];
/*        for i in l do
            l := subs(l, lcoeff(i) = 1)
        end_for; */
        l := map(l, dom::new);
        ret := {};
        for per in combinat::permute(nops(l)) do
            LL := _mult(l[i] $ i in per);
            p := poly(op(L), [X, Shift]);
            pp := poly(op(LL), [X, Shift]);
            eqs := {};
            for i in poly2list(pp) do
                eqs := eqs union { i[1] = expr(coeff(p, i[2])) };
            end_for;
            vars := [op(_union(indets(expr(i)) $ i in l) minus {X, Shift, q})];
            Sol := solve(eqs, [op(vars)], IgnoreSpecialCases);
        //  newvars := (indets(Sol) minus _union(indets(expr(i)) $ i in l)) minus {X, Shift, q};
        //  Sol := solve(eqs, [vars[i] $ i=1..nops(vars)-nops(newvars)], IgnoreSpecialCases);
            if Sol <> {} and domtype(Sol) = DOM_SET then
                newvars := (indets(Sol) minus {op(vars)}) minus {X, Shift, q};
                if nops(newvars) = 1 then
                    Sol := subs(Sol, op(newvars,1) = 1);
                end_if;
                ret := (map([l[i] $ i in per], (i) -> dom::new(subs(expr(i), op(Sol,1)))));
                break;
            elif Sol <> {} then
                ret := FAIL;
            end_if;
        end_for;
//    ), 3) then print("aborted"); return(FAIL) end_if;
    return(ret);
end_proc;


cgenerator := proc()
local word, word_int, A, i, first_time, s;
option escape;
begin
    if args(0)=0 then return(FAIL); end_if;
	A:=[args()]:
    for i from 1 to nops(A) do
        if domtype(A[i])=DOM_INT then
            A[i]:={$1..A[i]};
    elif testargs() and not (testtype(A[i], DOM_LIST) or testtype(A[i], DOM_SET)) then
	error("Expecting an integer, list, or set; got: ".expr2text(A[i]));
        end_if;
        if nops(A[i])=0 then
            return(FAIL);
        end_if;
    end_for;
    word_int := [1 $ args(0)];
    word     := [op(s,1) $ s in A];
    first_time:=TRUE;
    // Invariants:
    //  - word_int is a sequence of k integers with
    //		1 <= word_int[i] <= nops(A[i])
    //  - word is the corresponding element of A[1] x ... x A[k]
    //  - first_time is true if and only if it's the first run
    return(proc()
               local index;
           begin
               if first_time then
                   first_time:=FALSE;
                   return(word);
               end_if;
               for index from nops(A) downto 1 do
                   if word_int[index] < nops(A[index]) then
                       word_int[index] := word_int[index]+1;
                       word[index] := op(A[index], word_int[index]);
                       return(word);
                   else 
                       word_int[index] := 1;
                       word[index] := op(A[index],1);
                   end_if;
               end_for;
               return(FAIL);
           end_proc)
end_proc; // cgenerator

qgetfactors := proc(a)
local F, i, lc, n, cnt;
begin
  lc := lcoeff(a);
  cnt := content(numer(expr(a)),[X]);
  //print(cnt);
  F := factor(multcoeffs(a, 1/lc));
  f := Factored::factors(F);
  d := map(f, degree);
  l := map(f, ldegree);
  e := map(Factored::exponents(F), (x)->[i $ i=0..x]);
  [f, d, l, e, cnt];
end_proc;

qPetkovsek := proc(L : dom, Method="Muff")
local C0, Lc, Pa0, Paoo, Soo, a, ad, ae, af, ai, aiter, al, b, bd, be, bf, bi, biter, bl, 
c0, cc, f, i, ii, l, m, n, p, pa0, paoo, polrec, polsol, r, s, sol, tmp, u, x, zsol, alc, blc, BE;
//A,k,a,b,c,rec,i,j,l,r,s,p,ca,cb,polsol,polrec,m,n,f,x,u,zsol,sol,allzero,ratsol,qpowers,C0,S0, Soo, c0, tmp,
//Lc, Pa0, ad, ae, af, ai, aiter, al, bd, be, bf, bi, biter, bl, bops, cc, ii, pa0;
save z;
begin
  checktype := proc(c0, C0)
  local cc, tmp;
  begin
    //if not c0 in C0 and not -c0 in C0 then return(FALSE) else return(TRUE) end;
    if not c0 in C0 and not -c0 in C0 then  // traperror(subs(c0,q=1))=1 and not eval(subs(c0,q=1)) in CC0 then 
      for cc in C0 do
        tmp := normal(expand(cc/c0));
        //print(c0, cc, tmp);
        if testtype(tmp, Type::RatExpr(q)) then
          if tmp = q or tmp=-q or type(tmp) = "_power" or type(-tmp) = "_power" then 
            //print("*");
            return(TRUE);            
          end;
        end_if;
      end_for;
      return(FALSE);
    end_if;
    return(TRUE);
  end_proc;

  Dup2 := 0;
  Dup1:= 0;
  JJJ := 0;
  JJJJ := 0;
  L := dom::polyfy(L)[1];
  delete z;
  pa0 := dom::localTypeAt0(L);
  Pa0 := table(op(x,2) = {} $ x in pa0);
  for x in pa0 do
    if not iszero(x) then
      Pa0[op(x,2)] := Pa0[op(x,2)] union { normal(op(x,1)) }
    end_if
  end_for;
  paoo := dom::localTypeAtInifinity(L);
  Paoo := table(op(x,2) = {} $ x in paoo);
  for x in paoo do
    if not iszero(x) then
      Paoo[op(x,2)] := Paoo[op(x,2)] union { normal(op(x,1)) }
    end_if
  end_for;
  Soo := { op(map(paoo, op, 2)) };
  n := dom::order(L);
  Lc := [poly(coeff(L,i), [X]) $ i=0..n];
  a:=poly(tcoeff(L), [X]);
  b:=poly(subs(lcoeff(L), X=X*q^(-n+1)),[X]);
  [af,ad,al,ae,alc]:=(dom::qgetfactors(a,[X])); // generate a list of all combinations of factors of lowest coefficient polynomial
  [bf,bd,bl,be,blc]:=(dom::qgetfactors(b,[X])); // generate a list of all combinations of factors of highest coefficient polynomial
  sol:={};
  //print(Unquoted, "Insgesamt: ".expr2text(_mult(op(map(ae,_index,-1)), op(map(be,_index,-1))))." Möglichkeiten");
  ai := nops(af);
  bi := nops(bf);
  aiter := dom::cgenerator(op(ae));
  BE := be;
  while FAIL <> (ae := aiter()) do
    biter := dom::cgenerator(op(BE));
    while FAIL <> (be := biter()) do
      Dup1 := Dup1+1;
      if Method<>Traditional then
        if not -(_plus(ad[ii]*ae[ii] $ ii=1..ai)-_plus(bd[ii]*be[ii] $ ii=1..bi)) in Soo then 
          next;
        elif FAIL = (C0 := Pa0[_plus(al[ii]*ae[ii] $ ii=1..ai)-_plus(bl[ii]*be[ii] $ ii=1..bi)]) or type(C0) = "_index" then
          next;
        else
          a := _mult(tcoeff(af[ii])^ae[ii] $ ii=1..ai);
          b := _mult(tcoeff(bf[ii])^be[ii] $ ii=1..bi);
          c0 := normal(a/b);
          //print("Ja");
          if not checktype(c0, C0) and not checktype(normal(c0*alc/blc), C0) then
            next;
          end_if;
        end_if; 
      end_if;
      a := (_mult(af[ii]^ae[ii] $ ii=1..ai));
      b := (_mult(bf[ii]^be[ii] $ ii=1..bi));
      Dup2 := Dup2+1;
      f := [(Lc[l+1]*_mult(dom::qShift(a, m) $ m=0..l-1)*_mult(dom::qShift(b,m) $ m=l..n-1)) $ l=0..n]; // coefficient polynomials for recurrence of polynomial c(x)
      if 0 < (r := min(op(map(f, (x)->_if(iszero(x),infinity,ldegree(x, X)))))) then
        f := map(f, (x)->poly(map(poly2list(x), y -> [y[1],y[2]-r]),[X]));
      end_if;
      if Method <> hold(Polygon) then
        u := [ coeff(f[l+1],X,0) $ l=0..n];
        zsol:=solve(_plus(u[m+1]*z^m $ m=0..n),z,IgnoreSpecialCases);
        if type(zsol)="_union" then
          zsol := _union(_if(type(x)=DOM_SET,x,null())$ x in zsol);
        elif zsol = C_ or zsol = R_ or zsol = Q_ then
          error("This shouldn't happen!");
        end_if;  
      else
        zsol := [normal(x*lcoeff(b)/lcoeff(a)) $ x in Paoo[-(degree(a)-degree(b))]];
      end_if;
      //print(c0, zsol);
      for p in zsol do
        JJJ := JJJ + 1;
        //polrec:=dom::new(_plus(subs(z^s,p)*f[s+1]*Shift^s $ s=0..n)); // q-recurrence equation of polynomial c(x)
        //polrec:=new(dom, poly([[_plus(subs(z^s,p)*expr(f[s+1])), s] $ s=0..n], [Shift])); // q-recurrence equation of polynomial c(x)
        polrec:= [[expr(multcoeffs(f[s+1], p^s)), s] $ s=0..n]; // q-recurrence equation of polynomial c(x
        polrec:= poly(polrec, [Shift]); // q-recurrence equation of polynomial c(x)
        polrec:= new(dom, polrec);
        //print(polrec);
        bound := dom::qupperbound(polrec);
        
        if bound < 0 then next 
        elif bound > 0 and Method = hold(Polygon) then
          //print(bound, Method);
          polrec:=new(dom, poly([[_plus(p^s*expr(f[s+1]))*q^(-s*bound), s] $ s=0..n], [Shift])); // q-recurrence equation of polynomial c(x)
        end_if;
        polsol:=dom::qpolysols(polrec, TRUE, bound); // solve q-recurrence equation (polynomial solution);
        if nops(polsol)=0 then polsol := 0 else polsol := op(polsol, 1) end_if;
        if has(polsol,X) or polsol<>0 then
          //print(subs(z,p), Paoo[-(degree(a,[X])-degree(b,[X]))]);
          tmp := p*expr(a)/expr(b)*subs(polsol, X=q*X)/polsol;
//          return({tmp});
          //print(c0, p);
          sol:=sol union {tmp}; // create normal form z*a(x)/b(x)*c(q*x)/c(x)
        end_if;
      end_for;
    end_while;
  end_while;
  map(sol, (x)->dom::new(Shift-x));
end_proc;

//==================================================================
//* Domain Constructor
//------------------------------------------------------------------
begin // domain constructor,
  if {domtype(X), domtype(Shift), domtype(q)} <> {DOM_IDENT} then
    return("First two parameters must be identifiers");
  end_if;
  F := Dom::ExpressionField();
  // R := x -> error("Doof!");
end:

//==================================================================
// End of file ShiftOp.mu
//==================================================================
unalias (rep):

null():